/*
 * muse-client
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_client.h"
#include "muse_core_internal.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif
#define LOG_TAG "MUSED_CLIENT"

static GMutex mc_mutex;
static GHashTable *mc_table;
static int mc_table_id;

const char *UDS_files[MUSE_CHANNEL_MAX] = {MUSE_SOCK_FILE0, MUSE_SOCK_FILE1};

#ifdef MUSE_USE_CLIENT_SIGHANDLER
/* signal handler */
struct sigaction mc_int_old_action;
struct sigaction mc_abrt_old_action;
struct sigaction mc_segv_old_action;
struct sigaction mc_term_old_action;
struct sigaction mc_sys_old_action;

static gboolean _mc_table_remove_func(gpointer key, gpointer value, gpointer user_data);
static void _mc_sigaction(int signo, siginfo_t *siginfo, void *context);
static void _mc_constructor(void) __attribute__((constructor));
#endif

static int _mc_new(muse_channel_e channel);
static void _mc_table_new(void);
static gpointer _mc_get_fd_ptr(int sock_fd);

#ifdef MUSE_USE_CLIENT_SIGHANDLER
static gboolean _mc_table_remove_func(gpointer key, gpointer value, gpointer user_data)
{
	int sock_fd = GPOINTER_TO_INT(key);

	LOGW("close muse connection fd %d", sock_fd);

	muse_core_connection_close(sock_fd);

	return TRUE;
}

static void _mc_sigaction(int signo, siginfo_t *siginfo, void *context)
{
	struct sigaction *old_action = NULL;
	pid_t my_pid = getpid();

	LOGE("Enter - signo [%d], pid [%d]", signo, my_pid);

	/* close all opened fds */
	g_mutex_lock(&mc_mutex);

	if (mc_table)
		g_hash_table_foreach_remove(mc_table, _mc_table_remove_func, NULL);

	g_mutex_unlock(&mc_mutex);

	/* call old signal handler */
	switch (signo) {
	case SIGINT:
		old_action = &mc_int_old_action;
		break;
	case SIGABRT:
		old_action = &mc_abrt_old_action;
		break;
	case SIGSEGV:
		old_action = &mc_segv_old_action;
		break;
	case SIGTERM:
		old_action = &mc_term_old_action;
		break;
	case SIGSYS:
		old_action = &mc_sys_old_action;
		break;
	default:
		LOGE("unhandled signal %d", signo);
		return;
	}

	if (old_action->sa_sigaction)
		old_action->sa_sigaction(signo, siginfo, context);
	else
		sigaction(signo, old_action, NULL);

	raise(signo);

	LOGE("Leave");

	return;
}

static void _mc_constructor(void)
{
	struct sigaction mc_action;
	mc_action.sa_sigaction = _mc_sigaction;
	mc_action.sa_flags = SA_NOCLDSTOP | SA_SIGINFO;

	LOGI("Enter");

	sigemptyset(&mc_action.sa_mask);

	sigaction(SIGINT, &mc_action, &mc_int_old_action);
	sigaction(SIGABRT, &mc_action, &mc_abrt_old_action);
	sigaction(SIGSEGV, &mc_action, &mc_segv_old_action);
	sigaction(SIGTERM, &mc_action, &mc_term_old_action);
	sigaction(SIGSYS, &mc_action, &mc_sys_old_action);

	LOGI("Leave");

	return;
}
#endif

static int _mc_new(muse_channel_e channel)
{
	struct sockaddr_un address;
	int len, ret = MUSE_ERR;
	int sock_fd;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	LOGI("Enter");
	muse_return_val_if_fail(channel < MUSE_CHANNEL_MAX, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(muse_server_is_ready(), MM_ERROR_UNKNOWN);

	/* create socket */
	sock_fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (!muse_core_fd_is_valid(sock_fd)) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[socket failure] %d socket %s", errno, err_msg);
		muse_core_dump_fd_state(sock_fd);
		return ret;
	}
	LOGI("sockfd: %d", sock_fd);

	if (fcntl(sock_fd, F_SETFD, FD_CLOEXEC) < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("unable to set on ctrl socket fd %d: %s", sock_fd, err_msg);
		close(sock_fd);
		return ret;
	}
	LOGD("fcntl");

	memset(&address, 0, sizeof(address));
	address.sun_family = AF_UNIX;
	strncpy(address.sun_path, UDS_files[channel], sizeof(address.sun_path) - 1);
	len = sizeof(address);

	if (muse_core_set_nonblocking(sock_fd, false) != MM_ERROR_NONE)
		LOGE("Error - fd (%d) set blocking", sock_fd);

	if ((ret = connect(sock_fd, (struct sockaddr *)&address, len)) < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[Critical Error : %d] connect failure : %s", errno, err_msg);
		close(sock_fd);
		return ret;
	}

	LOGI("[%d] #%d guaranteed that connect was returned", sock_fd, mc_table_id);
	g_mutex_lock(&mc_mutex);
	g_hash_table_insert(mc_table, GINT_TO_POINTER(sock_fd), GINT_TO_POINTER(mc_table_id++));
	g_mutex_unlock(&mc_mutex);

	LOGI("Leave");
	return sock_fd;
}

static void _mc_table_new(void)
{
	muse_return_if_fail(muse_server_is_ready());

	g_mutex_lock(&mc_mutex);

	if (!mc_table) {
		mc_table_id = 1;
		mc_table = g_hash_table_new(g_direct_hash, g_direct_equal);
		LOGD("client table : %p", mc_table);
	}

	g_mutex_unlock(&mc_mutex);

	muse_core_create_fd_table();
}

static gpointer _mc_get_fd_ptr(int sock_fd)
{
	muse_return_val_if_fail(mc_table, NULL);

	return g_hash_table_lookup(mc_table, GINT_TO_POINTER(sock_fd));
}

int muse_client_new(void)
{
	_mc_table_new();
	return _mc_new(MUSE_CHANNEL_MSG);
}

int muse_client_new_data_ch(void)
{
	return _mc_new(MUSE_CHANNEL_DATA);
}

int muse_client_close(int sock_fd)
{
	int ret = MM_ERROR_NONE;

	ret = muse_core_connection_close(sock_fd);
	if (ret != MM_ERROR_NONE) {
		LOGE("close connection failed 0x%x", ret);
		return ret;
	}

	/* remove fd from table */
	g_mutex_lock(&mc_mutex);

	if (g_hash_table_remove(mc_table, GINT_TO_POINTER(sock_fd)))
		LOGD("success : remove fd %d from table", sock_fd);
	else
		LOGW("fail : remove fd %d from table[%p]", sock_fd, mc_table);

	g_mutex_unlock(&mc_mutex);

	return MM_ERROR_NONE;
}

int muse_client_ipc_push_data(int sock_fd, const char *data, int size, uint64_t data_id)
{
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	muse_recv_data_head_t header;
	int sended_len = 0;

	muse_return_val_if_fail(data, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(muse_core_fd_is_valid(sock_fd), MM_ERROR_INVALID_ARGUMENT);

	header.marker = MUSE_DATA_HEAD;
	header.size = size;
	header.id = data_id;

	LOGD("[%d] header.marker : %x header.size : %d header.id : %"G_GUINT64_FORMAT"",
			sock_fd, header.marker, header.size, header.id);

	if ((sended_len = send(sock_fd, &header, sizeof(muse_recv_data_head_t), 0)) < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] fail to send msg (%s) %d", sock_fd, err_msg, errno);
	}

	if ((sended_len += send(sock_fd, data, size, 0)) < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] fail to send msg (%s) %d", sock_fd, err_msg, errno);
	}

	return sended_len;
}

int muse_client_get_fd_id_value(int sock_fd)
{
	gpointer sock_id_ptr = _mc_get_fd_ptr(sock_fd);

	muse_return_val_if_fail(sock_id_ptr, MUSE_ERR);

	return GPOINTER_TO_INT(sock_id_ptr);
}

bool muse_client_check_fd_id_value(int sock_fd, int sock_id)
{
	gpointer sock_id_ptr = _mc_get_fd_ptr(sock_fd);

	muse_return_val_if_fail(sock_id_ptr, false);

	if (sock_id == GPOINTER_TO_INT(sock_id_ptr))
		return true;
	else
		return false;
}

int muse_client_get_module_index(const char *module_name, int *module_index)
{
	int idx = 0;
	char *str;
	char *hosts;
	char *host;
	char *ptr = NULL;
	bool module_name_matched = false;
	dictionary *mc_dict;

	muse_return_val_if_fail(module_name, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(module_index, MM_ERROR_INVALID_ARGUMENT);

	LOGD("Enter");

	mc_dict = iniparser_load(CONFFILE);
	muse_return_val_if_fail(mc_dict, PARSE_ERROR);

	str = iniparser_getstring(mc_dict, MUSE_HOST, NULL);
	if (!str) {
		LOGE("Error - iniparser_getstring (%s)", MUSE_HOST);
		iniparser_freedict(mc_dict);
		return PARSE_ERROR;
	}

	hosts = strdup(str);
	if (!hosts) {
		LOGE("Error - hosts allocation");
		iniparser_freedict(mc_dict);
		return PARSE_ERROR;
	}

	host = strtok_r(hosts, INI_PARSER_COMMA, &ptr);

	while (host) {
		g_strstrip(host);
		if (strncmp(module_name, host, strlen(module_name) + 1) == 0) {
			module_name_matched = true;
			*module_index = idx;
			break;
		}

		host = strtok_r(NULL, INI_PARSER_COMMA, &ptr);
		idx++;
	}

	free(hosts);

	iniparser_freedict(mc_dict);

	if (module_name_matched) {
		LOGD("Leave");
		return MM_ERROR_NONE;
	} else {
		LOGE("You have to check if [%s] is exact", module_name);
		*module_index = MUSE_ERR;
		LOGD("Leave");
		return MM_ERROR_INVALID_ARGUMENT;
	}
}
