/*
 * muse-core
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef __MUSE_CORE_H__
#define __MUSE_CORE_H__

#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>
#include <stdint.h>
#include <mm_error.h>

#define MUSE_URI_MAX_LENGTH		4096
#define MUSE_MSG_MAX_LENGTH		4096
#define MUSE_NUM_FD				4

#define MUSE_SERVER_READY		"/tmp/.muse_server_ready"
#define MUSE_DATA_HEAD			0xda1a6ead
#define MUSE_MSG_HEAD			0xda1a6eae
#define MUSE_PARAM_HANDLE		"handle"
#define MUSE_PARAM_RETURN		"ret"
#define MUSE_PARAM_EVENT		"event"

typedef struct muse_module *muse_module_h;

typedef enum {
	MUSE_PLAYER,
	MUSE_CAMERA,
	MUSE_RECORDER,
	MUSE_DRM,
	MUSE_MODULE_TYPE_MAX
} muse_core_api_module_e;

typedef enum {
	MUSE_TYPE_INT = 1,
	MUSE_TYPE_INT64,
	MUSE_TYPE_POINTER,
	MUSE_TYPE_DOUBLE,
	MUSE_TYPE_STRING,
	MUSE_TYPE_ARRAY,
	MUSE_TYPE_ANY,
	MUSE_TYPE_MAX
} muse_core_msg_type_e;

typedef enum {
	MUSE_MSG_PARSE_ERROR_NONE,
	MUSE_MSG_PARSE_ERROR_CONTINUE,
	MUSE_MSG_PARSE_ERROR_OTHER,
	MUSE_MSG_PARSE_ERROR_MAX
} muse_core_msg_parse_err_e;

typedef enum {
	MUSE_MODULE_COMMAND_INITIALIZE,
	MUSE_MODULE_COMMAND_SHUTDOWN,
	MUSE_MODULE_COMMAND_DEBUG_INFO_DUMP,
	MUSE_MODULE_COMMAND_CREATE_SERVER_ACK,
	MUSE_MODULE_COMMAND_RESOURCE_NOT_AVAILABLE,
	MUSE_MODULE_COMMAND_EXTERNAL_STORAGE_STATE_CHANGED,
	MUSE_MODULE_COMMAND_CREATE_CAUTION,
	MUSE_MODULE_COMMAND_MAX
} muse_module_command_e;

typedef enum {
	MUSE_CHANNEL_MSG,
	MUSE_CHANNEL_DATA,
	MUSE_CHANNEL_MAX
} muse_channel_e;

typedef struct muse_external_storage_info {
	int id;
	int state;
	const char *path;
} muse_external_storage_info_t;

extern const char *UDS_files[MUSE_CHANNEL_MAX];

int muse_core_connection_close(int sock_fd);
int muse_core_set_nonblocking(int fd, bool value);
int muse_core_set_socket_timeout(int sock_fd, int timeout_sec);
bool muse_core_fd_is_valid(int fd);
void muse_core_fd_close(int fd);

/* message */
int muse_core_msg_send(int sock_fd, const char *msg);
int muse_core_msg_recv(int sock_fd, char *msg);
int muse_core_msg_send_fd(int sock_fd, int *fds, const char *buf);
int muse_core_msg_recv_fd(int sock_fd, char *buf, int *out_fd);
char *muse_core_msg_new(int api, ...);
bool muse_core_msg_deserialize(const char *key, char *buf, int *parse_len, muse_core_msg_parse_err_e *err, muse_core_msg_type_e m_type, void *data);
void muse_core_msg_free(char *msg);
void *muse_core_msg_object_new(char *str, int *parse_len, muse_core_msg_parse_err_e *err);
bool muse_core_msg_object_get_value(const char *key, void *jobj, muse_core_msg_type_e m_type, void *data);
void muse_core_msg_object_free(void *msg_obj);

void muse_core_dump_fd_state(int fd);

void muse_core_create_fd_table(void);
void muse_core_remove_all_fd_table(void);
void muse_core_destroy_fd_table(void);
void muse_core_log_process_thread_info(int pid);
void muse_core_log_process_opened_fds(int pid);
void muse_core_log_process_cpu_memory(int pid);
void muse_core_log_file_list(const char *path);
void muse_core_log_cmd_info(char *cmd);
void muse_core_remove_symlink(const char *path);

bool muse_server_is_ready(void);

#ifdef __cplusplus
}
#endif
#endif	/*__MUSE_CORE_H__*/
