/*
 * muse-core
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef __MUSE_CORE_INTERNAL_H__
#define __MUSE_CORE_INTERNAL_H__

#ifdef __cplusplus
extern "C" {
#endif

#ifndef __USE_GNU
#define __USE_GNU
#endif

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "muse_core.h"
#include <dlog.h>
#include <fcntl.h>
#include <glib.h>
#include <gmodule.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <iniparser.h>

#define CONFFILE					SYSCONFDIR"/multimedia/mused.conf"
#define MUSE_HOST					"muse:hosts"
#define INI_PARSER_COMMA			","
#define MUSE_DATA_ROOT_PATH			TZ_SYS_DATA_PATH"/mused/"
#define MUSE_DEFAULT_PIDFILE		"/tmp/.muse-server.pid"
#define MUSE_SOCK_FILE0				"/tmp/.muse_server_socket"
#define MUSE_SOCK_FILE1				"/tmp/.muse_server_data_socket"
#define MUSE_LOG_PATH				"/var/log/mused"
#define MUSE_LOG_FILE				MUSE_LOG_PATH"/muselog"
#define MUSE_USE_LOG				"yes"

#define READ						0x02
#define PERSIST						0x10

#define MUSE_ERR					-1
#define SEND_FAIL					-1
#define RECV_FAIL					-1
#define MUSE_PARAM_MAX				10

#define PARSE_ERROR					-1

/* message */
#define MUSE_MSG_LEN				32
#define MUSE_MSG_LEN_MAX			128

/* dispatcher */
#define MUSE_DISPATCHER_MAX			512

#define MSG_KEY_API					"api"
#define MSG_KEY_MODULE_INDEX		"module"
#define MSG_KEY_MODULE_ADDR			"module_addr"

#define DISPATCHER					"dispatcher"
#define DISPATCHER_PTR				"dispatcher_pointer"
#define CMD_DISPATCHER				"cmd_dispatcher"
#define END_DELIM					'}'
#define END_DELIM_STR				" }"

#define muse_return_if_fail(expr) \
	if (!(expr)) { \
		LOGE("failed [%s]", #expr); \
		return; \
	}

#define muse_return_val_if_fail(expr, val) \
	if (!(expr)) { \
		LOGE("failed [%s]", #expr); \
		return (val); \
	}

#define MUSE_FREE(src) \
	do { \
		if (src) { \
			free(src); \
			src = NULL; \
		} \
	} while (0)

#define MUSE_G_FREE(src) \
	do { \
		if (src) { \
			g_free(src); \
			src = NULL; \
		} \
	} while (0)

typedef enum {
	API_CREATE,
	API_DESTROY,
	API_MAX = MUSE_DISPATCHER_MAX
} muse_api_type_e;

typedef struct muse_channel_info {
	GThread *thread;
	int sock_fd;
	int tbm_fd;
	union {
		GModule *dll_handle;
		struct {
			GQueue *data_queue;
			GMutex data_mutex;
			GCond data_cond;
		};
	};
} muse_channel_info_t;

typedef struct muse_module {
	muse_channel_info_t ch[MUSE_CHANNEL_MAX];
	char recv_msg[MUSE_MSG_MAX_LENGTH + 1];
	int msg_offset;
	int idx;
	int api;
	int pid;
	gpointer user_data;
	intptr_t handle;
	gboolean is_created;
	GMutex dispatch_lock;
} muse_module_t;

typedef struct muse_recv_data_head {
	unsigned int marker;
	uint64_t id;
	int size;
} muse_recv_data_head_t;

typedef struct muse_recv_data {
	muse_recv_data_head_t header;
	/* Dynamic allocated data area */
} muse_recv_data_t;

typedef enum {
	MUSE_MSG_TYPE_NORMAL,
	MUSE_MSG_TYPE_FDS
} muse_msg_type_e;

typedef struct muse_msg_info {
	unsigned int marker;
	muse_msg_type_e type;
	int size;
} muse_msg_info_t;

bool muse_core_msg_recv_len(int fd, char *buf, int msg_len);
void muse_core_update_fd_state(int fd);
void muse_core_get_cur_time(char *time_buf);

#ifdef __cplusplus
}
#endif
#endif	/*__MUSE_CORE_INTERNAL_H__*/
