/*
 * muse-core
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_core_internal.h"
#include <json.h>
#include <sys/mman.h>

#undef LOG_TAG
#define LOG_TAG						"MUSED_CORE"

#define MUSE_WATCHDOG_CHECK_PERIOD			10
#define MUSE_WATCHDOG_CHECK_COUNT			3
#define MUSE_WATCHDOG_TIMER_PERIOD			5

const char *msg_type[] = { "normal type", "fds type" };

static GMutex msg_lock;
static GMutex msg_ipc_lock;

static GMutex fd_state_lock;
static GHashTable *fd_state_table;

static int _muse_get_valid_fd_count(int *fds);
static void _muse_msg_json_set_error(muse_core_msg_parse_err_e *err, int jerr);
static json_object *_muse_msg_json_tokener_parse_len(const char *str, int *len, muse_core_msg_parse_err_e *err);
static gboolean _muse_msg_is_mem_ptr_valid(gpointer ptr);
static void _muse_msg_json_factory_args(json_object *jobj, va_list ap);
static json_object *_muse_msg_json_find_obj(json_object *jobj, const char *find_key);
static gboolean _muse_msg_json_get_obj_value(json_object *obj, muse_core_msg_type_e m_type, void *data);

static int _muse_get_valid_fd_count(int *fds)
{
	int idx;
	for (idx = 0; idx < MUSE_NUM_FD; idx++) {
		if (fds[idx] == MUSE_ERR)
			break;
	}
	return idx;
}

static void _muse_msg_json_set_error(muse_core_msg_parse_err_e *err, int jerr)
{
	if (err) {
		switch (jerr) {
		case json_tokener_success:
			*err = MUSE_MSG_PARSE_ERROR_NONE;
			break;
		case json_tokener_continue:
			*err = MUSE_MSG_PARSE_ERROR_CONTINUE;
			break;
		default:
			*err = MUSE_MSG_PARSE_ERROR_OTHER;
			break;
		}
	}
}

static json_object *_muse_msg_json_tokener_parse_len(const char *str, int *len, muse_core_msg_parse_err_e *err)
{
	struct json_tokener *tok = NULL;
	struct json_object *obj = NULL;
	int str_len = 0;

	muse_return_val_if_fail(str, NULL);

	tok = json_tokener_new();

	if (!tok) {
		LOGE("tokener creation error");
		goto out;
	}

	str_len = strlen(str);
	obj = json_tokener_parse_ex(tok, str, str_len);

	if (!obj) {
		LOGE("parsing error : [length : %d] %s", str_len, str);
		goto out;
	}

	if (len)
		*len = tok->char_offset;

	if (tok->err != json_tokener_success) {
		LOGE("Json Error(%d) : %s", tok->err, json_tokener_error_desc(tok->err));
		json_object_put(obj);
		obj = NULL;
	}
	_muse_msg_json_set_error(err, tok->err);

out:
	if (tok)
		json_tokener_free(tok);

	return obj;
}

static gboolean _muse_msg_is_mem_ptr_valid(gpointer ptr)
{
	size_t page_size = sysconf(_SC_PAGESIZE);
	gpointer base = (gpointer)((size_t)ptr / page_size * page_size);

	return msync(base, page_size, MS_ASYNC) == 0;
}

static void _muse_msg_json_factory_args(json_object *jobj, va_list ap)
{
	int type, len, idx;
	char *name;
	int *value;
	json_object *jarr;

	while ((type = va_arg(ap, int)) != 0) {
		name = va_arg(ap, char *);
		switch (type) {
		case MUSE_TYPE_INT:
			json_object_object_add(jobj, name, json_object_new_int(va_arg(ap, int32_t)));
			break;
		case MUSE_TYPE_INT64:
			json_object_object_add(jobj, name, json_object_new_int64(va_arg(ap, int64_t)));
			break;
		case MUSE_TYPE_POINTER:
			if (sizeof(intptr_t) == 8)
				json_object_object_add(jobj, name, json_object_new_int64(va_arg(ap, intptr_t)));
			else
				json_object_object_add(jobj, name, json_object_new_int(va_arg(ap, intptr_t)));
			break;
		case MUSE_TYPE_DOUBLE:
			json_object_object_add(jobj, name, json_object_new_double(va_arg(ap, double)));
			break;
		case MUSE_TYPE_STRING:
			json_object_object_add(jobj, name, json_object_new_string(va_arg(ap, char *)));
			break;
		case MUSE_TYPE_ARRAY:
			len = va_arg(ap, int);
			value = va_arg(ap, int *);
			jarr = json_object_new_array();

			for (idx = 0; idx < len; idx++)
				json_object_array_add(jarr, json_object_new_int(value[idx]));
			json_object_object_add(jobj, name, jarr);
			break;
		default:
			LOGE("Unexpected type");
		}
	}
}

static json_object *_muse_msg_json_find_obj(json_object *jobj, const char *find_key)
{
	size_t key_len = 0;

	muse_return_val_if_fail(jobj, NULL);

	muse_return_val_if_fail(find_key, NULL);

	key_len = strlen(find_key);

	json_object_object_foreach(jobj, key, val) {
		if (strlen(key) == key_len && !memcmp(key, find_key, key_len))
			return val;
	}

	return NULL;
}

static gboolean _muse_msg_json_get_obj_value(json_object *obj, muse_core_msg_type_e m_type, void *data)
{
	int j_type, idx, len;
	int *int_data;

	muse_return_val_if_fail(obj, FALSE);
	muse_return_val_if_fail(data, FALSE);

	j_type = json_object_get_type(obj);
	switch (j_type) {
	case json_type_null:
		LOGD("json_type_null");
		break;
	case json_type_double:
		*(double *)data = json_object_get_double(obj);
		break;
	case json_type_int:
		if (m_type == MUSE_TYPE_ANY || m_type == MUSE_TYPE_INT) {
			*(int32_t *)data = json_object_get_int(obj);
		} else if (m_type == MUSE_TYPE_INT64) {
			*(int64_t *)data = json_object_get_int64(obj);
		} else if (m_type == MUSE_TYPE_POINTER) {
			if (!_muse_msg_is_mem_ptr_valid(obj)) {
				LOGE("memory pointer is not valid");
				return FALSE;
			}
			if (sizeof(intptr_t) == 8)
				*(intptr_t *)data = json_object_get_int64(obj);
			else
				*(intptr_t *)data = json_object_get_int(obj);
		} else if (m_type == MUSE_TYPE_DOUBLE) {
			*(double *)data = json_object_get_double(obj);
		}
		break;
	case json_type_object:
		break;
	case json_type_string:
		strncpy((char *)data, json_object_get_string(obj), strlen(json_object_get_string(obj)));
		break;
	case json_type_array:
		int_data = (int *)data;
		len = json_object_array_length(obj);
		for (idx = 0; idx < len; idx++)
			int_data[idx] = json_object_get_int(json_object_array_get_idx(obj, idx));
		break;
	default:
		LOGE("The value (%d) of json type is invalid", j_type);
		break;
	}

	return TRUE;
}

void muse_core_log_cmd_info(char *cmd)
{
	FILE *fp;
	char buf[MUSE_MSG_LEN_MAX];

	muse_return_if_fail(cmd);

	fp = popen(cmd, "r");

	if (fp) {
		while (fgets(buf, MUSE_MSG_LEN_MAX, fp))
			LOGW("%s", buf);

		if (pclose(fp) == -1)
			LOGE("Fail to pclose");
	}
}

bool muse_server_is_ready(void)
{
	return access(MUSE_DEFAULT_PIDFILE, F_OK) == 0;
}

int muse_core_connection_close(int sock_fd)
{
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	if (!muse_core_fd_is_valid(sock_fd)) {
		if (sock_fd != STDIN_FILENO)
			LOGE("[%d] invalid socket", sock_fd);
		return MM_ERROR_INVALID_ARGUMENT;
	}

	if (shutdown(sock_fd, SHUT_RDWR) == MUSE_ERR) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] failed to shutdown [error %s %d]", sock_fd, err_msg, errno);
	}

	if (close(sock_fd) == MUSE_ERR) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] failed to close [error %s %d]", sock_fd, err_msg, errno);
	}

	muse_core_dump_fd_state(sock_fd);

	g_mutex_lock(&fd_state_lock);

	if (!g_hash_table_remove(fd_state_table, GINT_TO_POINTER(sock_fd)))
		LOGW("fail : remove fd %d from table[%p]", sock_fd, fd_state_table);

	g_mutex_unlock(&fd_state_lock);

	return MM_ERROR_NONE;
}

int muse_core_set_nonblocking(int fd, bool value)
{
	int flags = fcntl(fd, F_GETFL, NULL);
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	if (flags >= 0) {
		flags = value ? (flags | O_NONBLOCK) : (flags & ~O_NONBLOCK);
		if (fcntl(fd, F_SETFL, flags) == -1) {
			strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
			LOGE("fcntl (%d, F_SETFL, %d) [error %s %d]", fd, flags, err_msg, errno);
			return MM_ERROR_FILE_INTERNAL;
		} else {
			LOGD("fcntl (%d, F_SETFL)", fd);
		}
	} else {
		LOGE("failed to get flags for %d", fd);
		return MM_ERROR_FILE_INTERNAL;
	}

	return MM_ERROR_NONE;
}

int muse_core_set_socket_timeout(int sock_fd, int timeout_sec)
{
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	struct timeval tv;

	LOGD("Enter");

	tv.tv_sec = timeout_sec;
	tv.tv_usec = 0L;

	if (setsockopt(sock_fd, SOL_SOCKET, SO_SNDTIMEO, (void *)&tv, (socklen_t)sizeof(tv)) == MUSE_ERR) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] Failed to set socket send timeout option [error %s %d]", sock_fd, err_msg, errno);
		return MM_ERROR_UNKNOWN;
	}

	if (setsockopt(sock_fd, SOL_SOCKET, SO_RCVTIMEO, (void *)&tv, (socklen_t)sizeof(tv)) == MUSE_ERR) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] Failed to set socket recv timeout option [error %s %d]", sock_fd, err_msg, errno);
		return MM_ERROR_UNKNOWN;
	}

	LOGD("Leave");

	return MM_ERROR_NONE;
}

bool muse_core_fd_is_valid(int fd)
{
	muse_core_update_fd_state(fd);
	return (fcntl(fd, F_GETFL) != MUSE_ERR || errno != EBADF) && (fd > STDERR_FILENO);
}

void muse_core_fd_close(int fd)
{
	if (!muse_core_fd_is_valid(fd)) {
		LOGW("%d is invalid fd", fd);
		return;
	}

	close(fd);
}

/* message */
int muse_core_msg_send(int sock_fd, const char *msg)
{
	muse_return_val_if_fail(msg, MM_ERROR_INVALID_ARGUMENT);
	return muse_core_msg_send_fd(sock_fd, NULL, msg);
}

int muse_core_msg_send_fd(int sock_fd, int *fds, const char *buf)
{
	int ret = MM_ERROR_NONE;
	muse_msg_info_t msg_info = {0,};
	struct cmsghdr *cptr;
	struct msghdr msg;
	struct iovec iov;
	char data[CMSG_SPACE(sizeof(int) * MUSE_NUM_FD)];
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	int *fdptr;
	int fd_cnt = 0;

	muse_return_val_if_fail(buf, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(muse_core_fd_is_valid(sock_fd), MM_ERROR_INVALID_ARGUMENT);

	if (!fds)
		msg_info.type = MUSE_MSG_TYPE_NORMAL;
	else
		msg_info.type = MUSE_MSG_TYPE_FDS;

	msg_info.marker = MUSE_MSG_HEAD;
	msg_info.size = strlen(buf);

	g_mutex_lock(&msg_ipc_lock);

	ret = send(sock_fd, &msg_info, sizeof(muse_msg_info_t), 0);
	if (ret != sizeof(muse_msg_info_t)) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("msg info [type : %s size : %d] send failed : %d [error %s %d]",
				msg_type[msg_info.type], msg_info.size, ret, err_msg, errno);

		goto _MSG_SEND_DONE;
	}

	if (msg_info.type == MUSE_MSG_TYPE_NORMAL) {
		ret = send(sock_fd, buf, msg_info.size, 0);
		if (ret != (int)msg_info.size) {
			strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
			LOGE("[%s] send failed : %d [error %s %d]", buf, ret, err_msg, errno);
		}

		goto _MSG_SEND_DONE;
	}

	/* MUSE_MSG_TYPE_FDS */
	memset(&iov, 0, sizeof(iov));
	iov.iov_base = (void *)buf;
	iov.iov_len = msg_info.size;

	memset(&msg, 0, sizeof(msg));
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	msg.msg_control = data;
	msg.msg_controllen = sizeof(data);

	cptr = CMSG_FIRSTHDR(&msg);
	if (cptr) {
		fd_cnt = _muse_get_valid_fd_count(fds);
		cptr->cmsg_len = CMSG_LEN(sizeof(int) * fd_cnt);
		cptr->cmsg_level = SOL_SOCKET;
		cptr->cmsg_type = SCM_RIGHTS;
		fdptr = (int *)CMSG_DATA(cptr);

		memcpy(fdptr, fds, sizeof(int) * fd_cnt);

		/* the value of msg_controllen increases after memcpy so reassignes the orginal value */
		msg.msg_controllen = cptr->cmsg_len;
	}

	if ((ret = sendmsg(sock_fd, &msg, 0)) == SEND_FAIL) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] fail to send msg - [error %s %d]", sock_fd, err_msg, errno);
	}

_MSG_SEND_DONE:
	g_mutex_unlock(&msg_ipc_lock);

	return ret;
}

int muse_core_msg_recv(int sock_fd, char *msg)
{
	muse_return_val_if_fail(msg, MM_ERROR_INVALID_ARGUMENT);
	return muse_core_msg_recv_fd(sock_fd, msg, NULL);
}

int muse_core_msg_recv_fd(int sock_fd, char *buf, int *out_fd)
{
	int ret = 0;
	int pid;
	struct cmsghdr *cptr;
	struct msghdr msg;
	struct iovec iov;
	muse_msg_info_t msg_info = {0,};
	char data[CMSG_SPACE(sizeof(int) * MUSE_NUM_FD)];
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	muse_return_val_if_fail(buf, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(muse_core_fd_is_valid(sock_fd), MM_ERROR_INVALID_ARGUMENT);

	/* get the msg type and length */
	if (!muse_core_msg_recv_len(sock_fd, (char *)&msg_info, sizeof(muse_msg_info_t))) {
		pid = (int)getpid();
		LOGE("[pid : %d] [fd : %d] msg info receive failed", pid, sock_fd);
		muse_core_log_process_thread_info(pid);
		return RECV_FAIL;
	}

	if (msg_info.marker != MUSE_MSG_HEAD) {
		LOGE("invalid marker 0x%x", msg_info.marker);
		return RECV_FAIL;
	}

	if (msg_info.type == MUSE_MSG_TYPE_NORMAL) {
		if (!muse_core_msg_recv_len(sock_fd, buf, msg_info.size)) {
			LOGE("msg receive failed");
			return RECV_FAIL;
		}

		ret = msg_info.size;
	} else {
		memset(&iov, 0, sizeof(iov));
		iov.iov_base = (void *)buf;
		iov.iov_len = msg_info.size;

		memset(&msg, 0, sizeof(msg));
		msg.msg_name = NULL;
		msg.msg_namelen = 0;
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_control = data;
		msg.msg_controllen = sizeof(data);

		if ((ret = recvmsg(sock_fd, &msg, 0)) == RECV_FAIL) {
			strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
			LOGE("fail to receive msg [error %s %d]", err_msg, errno);
			return ret;
		}

		if (out_fd) {
			cptr = CMSG_FIRSTHDR(&msg);
			if (cptr)
				memcpy(out_fd, CMSG_DATA(cptr), cptr->cmsg_len - CMSG_LEN(0));
		}
	}

	buf[ret] = '\0';

	return ret;
}

char *muse_core_msg_new(int api, ...)
{
	json_object *jobj;
	const char *jsonMsg;
	char *sndMsg;
	va_list ap;

	jobj = json_object_new_object();

	muse_return_val_if_fail(jobj, NULL);

	json_object_object_add(jobj, MSG_KEY_API, json_object_new_int(api));

	va_start(ap, api);
	_muse_msg_json_factory_args(jobj, ap);
	va_end(ap);

	jsonMsg = json_object_to_json_string(jobj);
	sndMsg = g_strdup(jsonMsg);

	if (api == API_CREATE)
		SECURE_LOGD("%s", sndMsg);

	json_object_put(jobj);

	return sndMsg;
}

bool muse_core_msg_deserialize(const char *key, char *buf, int *parse_len, muse_core_msg_parse_err_e *err, muse_core_msg_type_e m_type, void *data)
{
	json_object *obj = NULL, *jobj = NULL;
	bool ret = false;

	muse_return_val_if_fail(key, false);
	muse_return_val_if_fail(buf, false);
	muse_return_val_if_fail(m_type >= MUSE_TYPE_INT && m_type < MUSE_TYPE_MAX, false);
	muse_return_val_if_fail(data, false);

	g_mutex_lock(&msg_lock);

	jobj = _muse_msg_json_tokener_parse_len(buf, parse_len, err);
	if (!jobj) {
		LOGE("jobj is NULL");
		g_mutex_unlock(&msg_lock);
		return false;
	}

	obj = _muse_msg_json_find_obj(jobj, key);
	if (!obj) {
		LOGE("\"%s\" key is not founded", key);
		json_object_put(jobj);
		g_mutex_unlock(&msg_lock);
		return false;
	}

	ret = (bool)_muse_msg_json_get_obj_value(obj, m_type, data);

	json_object_put(jobj);

	g_mutex_unlock(&msg_lock);

	return true;
}

void muse_core_msg_free(char *msg)
{
	MUSE_G_FREE(msg);
}

void *muse_core_msg_object_new(char *str, int *parse_len, muse_core_msg_parse_err_e *err)
{
	void *jobj = NULL;

	muse_return_val_if_fail(str, NULL);

	g_mutex_lock(&msg_lock);

	jobj = (void *)_muse_msg_json_tokener_parse_len(str, parse_len, err);

	g_mutex_unlock(&msg_lock);

	return jobj;
}

bool muse_core_msg_object_get_value(const char *key, void *jobj, muse_core_msg_type_e m_type, void *data)
{
	json_object *obj;
	bool ret = false;

	muse_return_val_if_fail(key, false);
	muse_return_val_if_fail(jobj, false);
	muse_return_val_if_fail(data, false);
	muse_return_val_if_fail(m_type >= MUSE_TYPE_INT && m_type < MUSE_TYPE_MAX, false);

	g_mutex_lock(&msg_lock);

	obj = _muse_msg_json_find_obj((json_object *)jobj, key);
	if (!obj) {
		LOGE("\"%s\" key is not found", key);
		g_mutex_unlock(&msg_lock);
		return false;
	}

	ret = (bool)_muse_msg_json_get_obj_value(obj, m_type, data);

	g_mutex_unlock(&msg_lock);

	return ret;
}

void muse_core_msg_object_free(void *jobj)
{
	muse_return_if_fail(jobj);

	g_mutex_lock(&msg_lock);

	json_object_put((json_object *)jobj);

	g_mutex_unlock(&msg_lock);
}

bool muse_core_msg_recv_len(int fd, char *buf, int msg_len)
{
	int offset = 0;
	int recv_len = 0;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	if (!muse_core_fd_is_valid(fd) || !buf || msg_len <= 0) {
		LOGE("invalid param : fd %d, buf %p, msg_len %d", fd, buf, msg_len);
		return false;
	}

	do {
		recv_len = recv(fd, buf + offset, msg_len - offset, 0);
		if (recv_len < 0) {
			strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
			LOGE("[fd %d] recv failed [error %s %d]", fd, err_msg, errno);
			return false;
		} else if (recv_len == 0) {
			LOGD("EOF");
			return false;
		}

		offset += recv_len;
	} while (offset < msg_len);

	if (offset != msg_len) {
		LOGE("invalid length received : try %d -> result %d", msg_len, offset);
		return false;
	}

	return true;
}

void muse_core_dump_fd_state(int fd)
{
	g_mutex_lock(&fd_state_lock);
	LOGI("[%d] %s", fd, (char *)g_hash_table_lookup(fd_state_table, GINT_TO_POINTER(fd)));
	g_mutex_unlock(&fd_state_lock);
}

void muse_core_update_fd_state(int fd)
{
	char atime[MUSE_MSG_LEN_MAX] = {'\0',};

	g_mutex_lock(&fd_state_lock);

	if (!fd_state_table)
		fd_state_table = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, g_free);

	muse_core_get_cur_time(atime);
	g_hash_table_insert(fd_state_table, GINT_TO_POINTER(fd), g_strdup(atime));

	g_mutex_unlock(&fd_state_lock);
}

void muse_core_create_fd_table(void)
{
	g_mutex_lock(&fd_state_lock);

	if (fd_state_table) {
		LOGW("fd state table [%p] is already created", fd_state_table);
		goto out;
	}

	fd_state_table = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, g_free);
	if (!fd_state_table)
		LOGE("fail to create new fd state table");
out:
	g_mutex_unlock(&fd_state_lock);
}

void muse_core_remove_all_fd_table(void)
{
	g_mutex_lock(&fd_state_lock);
	g_hash_table_remove_all(fd_state_table);
	g_mutex_unlock(&fd_state_lock);
}

void muse_core_destroy_fd_table(void)
{
	g_mutex_lock(&fd_state_lock);
	g_hash_table_destroy(fd_state_table);
	fd_state_table = NULL;
	g_mutex_unlock(&fd_state_lock);
}

void muse_core_get_cur_time(char *time_buf)
{
	char cur_time[MUSE_MSG_LEN];
	struct tm newtime;
	struct timeval tv;

	muse_return_if_fail(time_buf);

	gettimeofday(&tv, NULL);
	strftime(cur_time, MUSE_MSG_LEN, "%m-%d %H:%M:%S", localtime_r(&(tv.tv_sec), &newtime));

	snprintf(time_buf, MUSE_MSG_LEN_MAX, "%s.%03ld", cur_time, tv.tv_usec / 1000);
}

void muse_core_log_process_thread_info(int pid)
{
	char cmd[MUSE_MSG_LEN_MAX];

	snprintf(cmd, sizeof(cmd), "/bin/cat /proc/%d/status | /bin/grep Threads", pid);
	muse_core_log_cmd_info(cmd);
}

void muse_core_log_process_opened_fds(int pid)
{
	char dir_path[MUSE_MSG_LEN_MAX];
	struct dirent *de = NULL;
	DIR *dr = NULL;

	snprintf(dir_path, sizeof(dir_path), "/proc/%d/fd", pid); /* /proc/%d/fd is directory */
	dr = opendir(dir_path);
	muse_return_if_fail(dr)

	LOGI("directory path : %s", dir_path);

	while ((de = readdir(dr)) != NULL)
		LOGW("%s\n", de->d_name);

	closedir(dr);
}

void muse_core_log_process_cpu_memory(int pid)
{
	char cmd[MUSE_MSG_LEN_MAX];

	snprintf(cmd, sizeof(cmd), "/bin/ps -Lo pcpu,pmem,tid,comm -p %d", pid);
	muse_core_log_cmd_info(cmd);
}

void muse_core_log_file_list(const char *path)
{
	char cmd[MUSE_MSG_LEN_MAX];

	snprintf(cmd, sizeof(cmd), "/bin/ls -alt %s", path);
	muse_core_log_cmd_info(cmd);
}

void muse_core_remove_symlink(const char *path)
{
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	char *r = NULL;

	muse_return_if_fail(path);

	r = realpath(path, NULL);
	if (r) {
		if (strncmp(r, path, strlen(path) + 1) != 0) {
			LOGW("symbolic link exists [%s] -> [%s]", path, r);
			unlink(path);
		}
		free(r);
	} else {
		if (errno != ENOENT) {
			strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
			LOGE("realpath failed [error %s %d]", err_msg, errno);
		}
	}
}

