/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"

void muse_server_set_dispatch_timeout(muse_module_h m, int api, int timeout)
{
	muse_return_if_fail(m);
	ms_module_set_timeout(m->idx, api, timeout);
}

int muse_server_get_dispatch_timeout(muse_module_h m, int api)
{
	muse_return_val_if_fail(m, MUSE_ERR);
	return ms_module_get_timeout(m->idx, api);
}

/* module */
int muse_server_module_set_value(muse_module_h m, const char *name, int value)
{
	muse_return_val_if_fail(m, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(name, MM_ERROR_INVALID_ARGUMENT);
	ms_module_set_dllsym_value(m->idx, name, GINT_TO_POINTER(value));
	return MM_ERROR_NONE;
}

int muse_server_module_get_value(muse_module_h m, const char *name, int *value)
{
	int ret = MM_ERROR_NONE;
	gpointer ptr;

	muse_return_val_if_fail(m, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(name, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(value, MM_ERROR_INVALID_ARGUMENT);

	ret = ms_module_get_dllsym_value(m->idx, name, &ptr);
	if (ret == MM_ERROR_NONE)
		*value = GPOINTER_TO_INT(ptr);
	return ret;
}

int muse_server_module_get_msg_fd(muse_module_h m)
{
	muse_return_val_if_fail(m, MM_ERROR_INVALID_ARGUMENT);
	return m->ch[MUSE_CHANNEL_MSG].sock_fd;
}

int muse_server_module_get_data_fd(muse_module_h m)
{
	muse_return_val_if_fail(m, MM_ERROR_INVALID_ARGUMENT);
	return m->ch[MUSE_CHANNEL_DATA].sock_fd;
}

char *muse_server_module_get_msg(muse_module_h m)
{
	muse_return_val_if_fail(m, NULL);
	return (m->recv_msg + m->msg_offset);
}

void muse_server_set_user_data(muse_module_h m, void *data)
{
	muse_return_if_fail(m);
	m->user_data = data;
}

void *muse_server_get_user_data(muse_module_h m)
{
	muse_return_val_if_fail(m, NULL);
	return m->user_data;
}

bool muse_server_module_is_valid(muse_module_h m)
{
	muse_return_val_if_fail(m, false);
	muse_return_val_if_fail(m->pid > 0, false);
	muse_return_val_if_fail(muse_core_fd_is_valid(m->ch[MUSE_CHANNEL_MSG].sock_fd), false);

	return true;
}

/* ipc */
int muse_server_ipc_get_bufmgr(tbm_bufmgr *bufmgr)
{
	LOGD("Enter");

	muse_return_val_if_fail(ms_get_instance(), MM_ERROR_UNKNOWN);
	muse_return_val_if_fail(bufmgr, MM_ERROR_INVALID_ARGUMENT);

	if (!ms_get_instance()->bufmgr)
		muse_return_val_if_fail(ms_init_bufmgr(), MM_ERROR_INVALID_ARGUMENT);

	LOGD("bufmgr: %p", ms_get_instance()->bufmgr);
	*bufmgr = ms_get_instance()->bufmgr;

	LOGD("Leave");
	return MM_ERROR_NONE;
}

int muse_server_ipc_get_gdbus_connection(GDBusConnection **gdbus)
{
	LOGD("Enter");

	muse_return_val_if_fail(gdbus, MM_ERROR_INVALID_ARGUMENT);
	if (!ms_system_get_gdbus(gdbus))
		return MM_ERROR_UNKNOWN;

	LOGD("gdbus : %p", *gdbus);
	return MM_ERROR_NONE;
}

char *muse_server_ipc_get_data(muse_module_h m)
{
	muse_channel_info_t *ch = NULL;
	char *raw_data = NULL;

	gint64 end_time = g_get_monotonic_time() + 100 * G_TIME_SPAN_MILLISECOND;
	muse_return_val_if_fail(m, NULL);

	ch = &m->ch[MUSE_CHANNEL_DATA];
	muse_return_val_if_fail(ch->data_queue, NULL);

	g_mutex_lock(&ch->data_mutex);
	while (!(raw_data = (char *)g_queue_pop_head(ch->data_queue))) {
		if (!g_cond_wait_until(&ch->data_cond, &ch->data_mutex, end_time)) {
			LOGE("[%d] timeout has passed", ch->sock_fd);
			break;
		}
	}
	g_mutex_unlock(&ch->data_mutex);

	if (!raw_data) {
		LOGE("NULL raw_data");
		return NULL;
	}

	return (raw_data + sizeof(muse_recv_data_head_t));
}

void muse_server_ipc_delete_data(char *data)
{
	char *raw_data = NULL;

	muse_return_if_fail(data);

	raw_data = data - sizeof(muse_recv_data_head_t);

	g_free(raw_data);
}

bool muse_server_ipc_get_data_info(char *data, uint64_t *data_id, int *size)
{
	muse_recv_data_t *qdata = NULL;

	muse_return_val_if_fail(data, false);

	qdata = (muse_recv_data_t *)(data - sizeof(muse_recv_data_head_t));
	if (qdata && qdata->header.marker == MUSE_DATA_HEAD) {
		if (data_id)
			*data_id = qdata->header.id;
		if (size)
			*size = qdata->header.size;

		return true;
	}

	return false;
}

intptr_t muse_server_ipc_get_handle(muse_module_h m)
{
	muse_return_val_if_fail(m, 0);
	muse_return_val_if_fail(m->handle, 0);

	return m->handle;
}

int muse_server_ipc_set_handle(muse_module_h m, intptr_t handle)
{
	muse_return_val_if_fail(m, MM_ERROR_INVALID_ARGUMENT);

	m->handle = handle;
	return MM_ERROR_NONE;
}

int muse_server_ipc_get_fd(muse_module_h m, int *fd)
{
	muse_return_val_if_fail(m, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(fd, MM_ERROR_INVALID_ARGUMENT);

	*fd = m->ch[MUSE_CHANNEL_MSG].tbm_fd;
	return MM_ERROR_NONE;
}

int muse_server_ipc_set_fd(muse_module_h m, int fd)
{
	muse_return_val_if_fail(m, MM_ERROR_INVALID_ARGUMENT);

	if (!muse_core_fd_is_valid(fd)) {
		LOGE("invalid fd: %d", fd);
		return MM_ERROR_INVALID_ARGUMENT;
	}

	m->ch[MUSE_CHANNEL_MSG].tbm_fd = fd;
	return MM_ERROR_NONE;
}

int muse_server_ipc_get_client_pid(muse_module_h m)
{
	muse_return_val_if_fail(m, MUSE_ERR);
	return m->pid;
}

/* security */
bool muse_server_security_check_cynara(int fd, const char *privilege)
{
	muse_return_val_if_fail(ms_get_instance(), false);
	muse_return_val_if_fail(ms_get_instance()->security, false);

	return ms_security_is_privilege_granted(ms_get_instance()->security, fd, privilege);
}

unsigned int muse_server_get_atomic_uint(void)
{
	static guint atom = 0;

	g_atomic_int_inc(&atom);

	return atom;
}

int muse_server_get_platform_info(const char *key, bool *value)
{
	return ms_system_get_platform_info(key, value);
}

int muse_server_get_module_instance_count(int idx)
{
	ms_connection_t *connection = NULL;

	muse_return_val_if_fail(ms_get_instance(), MUSE_ERR);

	connection = ms_get_instance()->connection;
	muse_return_val_if_fail(connection, MUSE_ERR);

	return connection->instance_count[idx];
}

bool muse_server_is_on_demand(void)
{
	ms_config_t *conf = NULL;

	muse_return_val_if_fail(ms_get_instance(), FALSE);

	conf = ms_get_instance()->conf;
	muse_return_val_if_fail(conf, FALSE);

	return conf->is_on_demand;
}

int main(int argc, char **argv)
{
#ifdef MUSE_TTRACE_LOG
	trace_begin("MUSE:START");
#endif

	if (argc > 1)
		ms_parse_params(argc, argv);

	ms_init(argv);

	ms_run();

	return ms_deinit();
}
