/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"

static char *_ms_config_get_str(dictionary *d, const char *key, char *def);
static int _ms_config_get_int(dictionary *d, const char *key, int def);
static int _ms_config_parser(ms_config_t *conf);
static ms_config_t *_ms_config_get_instance(void);

static char *_ms_config_get_str(dictionary *d, const char *key, char *def)
{
	char *str = NULL;

	muse_return_val_if_fail(d, NULL);
	muse_return_val_if_fail(key, NULL);

	str = iniparser_getstring(d, key, def);
	if (!str) {
		LOGE("Error - iniparser_getstring (%s)", key);
		return NULL;
	}

	return strdup(str);
}

static int _ms_config_get_int(dictionary *d, const char *key, int def)
{
	muse_return_val_if_fail(d, PARSE_ERROR);
	muse_return_val_if_fail(key, PARSE_ERROR);

	return iniparser_getint(d, key, def);
}

static int _ms_config_parser(ms_config_t *conf)
{
	int host_idx;
	char *hosts = NULL;
	char *host = NULL;
	char *str = NULL;
	char *ptr = NULL;
	char *key = NULL;
	char gst_param_key[MUSE_MSG_LEN_MAX];

	muse_return_val_if_fail(conf, PARSE_ERROR);

	conf->muse_dict = iniparser_load(CONFFILE);
	muse_return_val_if_fail(conf->muse_dict, PARSE_ERROR);

	hosts = _ms_config_get_str(conf->muse_dict, MUSE_HOST, NULL);
	if (!hosts) {
		LOGE("Error - hosts allocation");
		goto out;
	}

	str = _ms_config_get_str(conf->muse_dict, MUSE_LOG, NULL);
	if (str) {
		if (strncmp(str, MUSE_USE_LOG, strlen(MUSE_USE_LOG) + 1) == 0)
			conf->log_enabled = TRUE;
		else
			conf->log_enabled = FALSE;
		free(str);
	} else {
		LOGE("[%s] get string error", MUSE_LOG);
		conf->log_enabled = FALSE;
	}

	conf->lockfile = _ms_config_get_str(conf->muse_dict, MUSE_LOCK, NULL);
	if (!conf->lockfile) {
		LOGE("Error - lockfile allocation");
		goto out;
	}

	conf->min_timeout = _ms_config_get_int(conf->muse_dict, MUSE_WATCHDOG_MIN_TIMEOUT, DEFAULT_WATCHDOG_MIN_TIMEOUT);

	conf->max_timeout = _ms_config_get_int(conf->muse_dict, MUSE_WATCHDOG_MAX_TIMEOUT, DEFAULT_WATCHDOG_MAX_TIMEOUT);

	conf->memory_threshold = _ms_config_get_int(conf->muse_dict, MUSE_MEMORY_THRESHOLD, DEFAULT_MEMORY_LEAK_THRESHOLD);

	conf->is_watch_external_storage_enabled = (gboolean)_ms_config_get_int(conf->muse_dict, MUSE_WATCH_EXTERNAL_STORAGE, FALSE);

	conf->is_on_demand = (gboolean)_ms_config_get_int(conf->muse_dict, MUSE_ON_DEMAND_LAUNCH, FALSE);

	conf->max_idle_time = _ms_config_get_int(conf->muse_dict, MUSE_ON_DEMAND_MAX_IDLE_TIME, DEFAULT_ON_DEMAND_MAX_IDLE_TIME);

	conf->max_idle_memory = _ms_config_get_int(conf->muse_dict, MUSE_ON_DEMAND_MAX_IDLE_MEMORY, DEFAULT_ON_DEMAND_MAX_IDLE_MEMORY);

	for (conf->gst_param_cnt = 0; conf->gst_param_cnt < MUSE_PARAM_MAX; conf->gst_param_cnt++) {
		memset(gst_param_key, 0, MUSE_MSG_LEN_MAX);
		snprintf(gst_param_key, strlen(MUSE_GST) + 3, "%s%d", MUSE_GST, conf->gst_param_cnt + 1);

		conf->gst_param_str[conf->gst_param_cnt] = _ms_config_get_str(conf->muse_dict, gst_param_key, NULL);

		if (!conf->gst_param_str[conf->gst_param_cnt] || strlen(conf->gst_param_str[conf->gst_param_cnt]) == 0) {
			LOGD("completed to update gst_param #: %d", conf->gst_param_cnt);
			break;
		}
		LOGD("gstparam%d: %s", conf->gst_param_cnt + 1, conf->gst_param_str[conf->gst_param_cnt]);
	}

	conf->host_cnt = 0;
	host = strtok_r(hosts, INI_PARSER_COMMA, &ptr);

	while (host) {
		host_idx = conf->host_cnt;
		g_strstrip(host);
		conf->host[host_idx] = strdup(host);
		LOGD("host: %s", conf->host[host_idx]);
		key = (char *)malloc(MUSE_MSG_LEN_MAX);
		if (!key) {
			LOGE("Error - null key");
			goto out;
		}

		conf->host_infos[host_idx] = (host_info_t *) calloc(1, sizeof(host_info_t));
		if (!conf->host_infos[host_idx]) {
			LOGE("Error - null type");
			goto out;
		}

		/* path */
		snprintf(key, MUSE_MSG_LEN_MAX, "%s%s%s", host, INI_PARSER_COLON, PATH);
		conf->host_infos[host_idx]->path = _ms_config_get_str(conf->muse_dict, key, NULL);
		if (!conf->host_infos[host_idx]->path) {
			LOGE("Error - null path");
			goto out;
		}

		LOGD("[%d] %s", host_idx, conf->host_infos[host_idx]->path);

		/* preloaded */
		snprintf(key, MUSE_MSG_LEN_MAX, "%s%s%s", host, INI_PARSER_COLON, PRELOADED);
		conf->host_infos[host_idx]->preloaded = _ms_config_get_str(conf->muse_dict, key, NULL);
		if (!conf->host_infos[host_idx]->preloaded) {
			LOGE("Error - null preloaded");
			goto out;
		}

		/* timeout */
		snprintf(key, MUSE_MSG_LEN_MAX, "%s%s%s", host, INI_PARSER_COLON, TIMEOUT);
		conf->host_infos[host_idx]->timeout = _ms_config_get_int(conf->muse_dict, key, DEFAULT_DISPATCH_WATCHDOG_TIMEOUT);
		if (conf->host_infos[host_idx]->timeout < conf->min_timeout || conf->host_infos[host_idx]->timeout > conf->max_timeout) {
			LOGE("Error - set timeout value [%d] ( %d ~ %d s )", conf->host_infos[host_idx]->timeout, conf->min_timeout, conf->max_timeout);
			goto out;
		}

		/* disable_watchdog_api */
		snprintf(key, MUSE_MSG_LEN_MAX, "%s%s%s", host, INI_PARSER_COLON, DISABLE_WATCHDOG_API_INDEX);
		conf->host_infos[host_idx]->disable_watchdog_api = _ms_config_get_int(conf->muse_dict, key, DEFAULT_WATCHDOG_DISABLED_API_INDEX);

		/* max instance */
		snprintf(key, MUSE_MSG_LEN_MAX, "%s%s%s", host, INI_PARSER_COLON, MAX_INSTANCE);
		conf->host_infos[host_idx]->max_instance = _ms_config_get_int(conf->muse_dict, key, DEFAULT_MAX_INSTANCE);

		/* caution instance */
		snprintf(key, MUSE_MSG_LEN_MAX, "%s%s%s", host, INI_PARSER_COLON, CAUTION_INSTANCE);
		conf->host_infos[host_idx]->caution_instance = _ms_config_get_int(conf->muse_dict, key, DEFAULT_CAUTION_INSTANCE);

		host = strtok_r(NULL, INI_PARSER_COMMA, &ptr);
		conf->host_cnt++;
		free(key);
	}

	free(hosts);
	return MM_ERROR_NONE;

out:
	ms_config_deinit(conf);
	MUSE_FREE(key);
	MUSE_FREE(hosts);
	return PARSE_ERROR;
}

static ms_config_t *_ms_config_get_instance(void)
{
	muse_return_val_if_fail(ms_get_instance(), NULL);

	return ms_get_instance()->conf;
}

void ms_config_init(ms_config_t *conf)
{
	muse_return_if_fail(conf);

	if (_ms_config_parser(conf) != MM_ERROR_NONE)
		LOGE("parser() error");
}

void ms_config_deinit(ms_config_t *conf)
{
	int idx = 0;

	muse_return_if_fail(conf);
	muse_return_if_fail(conf->host_infos);

	for (idx = 0; idx <= conf->host_cnt; idx++) {
		MUSE_FREE(conf->host_infos[conf->host_cnt]->path);
		MUSE_FREE(conf->host_infos[conf->host_cnt]->preloaded);
		MUSE_FREE(conf->host_infos[conf->host_cnt]);
		MUSE_FREE(conf->host[idx]);
	}

	MUSE_FREE(conf->lockfile);

	for (idx = 0; idx <= conf->gst_param_cnt; idx++)
		MUSE_FREE(conf->gst_param_str[idx]);

	free(conf);
}

char *ms_config_get_host_name(int idx)
{
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_val_if_fail(conf, NULL);
	muse_return_val_if_fail(conf->host, NULL);

	return conf->host[idx];
}

int ms_config_get_host_cnt(void)
{
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_val_if_fail(conf, MUSE_ERR);
	return conf->host_cnt;
}

int ms_config_get_gst_param_cnt(void)
{
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->gst_param_cnt;
}

char *ms_config_get_gst_param_str(int idx)
{
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_val_if_fail(conf, NULL);
	muse_return_val_if_fail(conf->gst_param_str[idx], NULL);
	return conf->gst_param_str[idx];
}

char *ms_config_get_path(int idx)
{
	char *path = NULL;
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_val_if_fail(conf, NULL);

	path = conf->host_infos[idx]->path;
	muse_return_val_if_fail(path, NULL);

	LOGD("%s", path);
	return path;
}

char *ms_config_get_preloaded_value(int idx)
{
	char *value = NULL;
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_val_if_fail(conf, NULL);

	value = conf->host_infos[idx]->preloaded;
	muse_return_val_if_fail(value, NULL);

	return value;
}

char *ms_config_get_lockfile(void)
{
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_val_if_fail(conf, NULL);

	return conf->lockfile;
}

gboolean ms_config_is_log_enabled(void)
{
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_val_if_fail(conf, FALSE);

	return conf->log_enabled;
}

void ms_config_remove_lockfile(void)
{
	ms_config_t *conf = _ms_config_get_instance();

	muse_return_if_fail(conf);

	if (remove(conf->lockfile) == -1)
		LOGE("remove %s failed [errno : %d]", conf->lockfile, errno);
}

int ms_config_get_min_timeout(void)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->min_timeout;
}

int ms_config_get_max_timeout(void)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->max_timeout;
}

int ms_config_get_memory_threshold(void)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->memory_threshold;
}

gboolean ms_config_is_on_demand(void)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, FALSE);

	return conf->is_on_demand;
}

int ms_config_get_max_idle_time(void)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->max_idle_time;
}

int ms_config_get_timeout(int module_idx)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->host_infos[module_idx]->timeout;
}

int ms_config_get_disable_watchdog_api(int module_idx)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->host_infos[module_idx]->disable_watchdog_api;
}

int ms_config_get_max_instance(int module_idx)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->host_infos[module_idx]->max_instance;
}

int ms_config_get_caution_instance(int module_idx)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, MUSE_ERR);

	return conf->host_infos[module_idx]->caution_instance;
}

gboolean ms_config_is_external_storage_enabled(void)
{
	ms_config_t *conf = _ms_config_get_instance();
	muse_return_val_if_fail(conf, FALSE);

	return conf->is_watch_external_storage_enabled;
}

