/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_core_internal.h"
#include "muse_server_private.h"

#define MSG_THREAD_NAME						"msg"
#define DATA_THREAD_NAME					"data"

#define DATA_WORKER_QDATA_MAX_SIZE			(3840 * 2160 * 4) /* UHD BGRA8888 */
#define UNLIMITED_INSTANCE					-1

static void _ms_ipc_data_ch_init(muse_module_h m);
static void _ms_ipc_data_ch_deinit(muse_module_h m);
static void _ms_ipc_module_data_ch_cleanup(muse_module_h m);
static void _ms_ipc_module_msg_ch_cleanup(muse_module_h m);
static void _ms_ipc_module_cleanup(muse_module_h m);
static gboolean _ms_ipc_module_instance_creation_is_allowed(int module_idx);
static gboolean _ms_ipc_get_module_idx(muse_module_h m, void *jobj);
static gboolean _ms_ipc_dispatch_create(muse_module_h m, void *jobj);
static gboolean _ms_ipc_dispatch_destroy(muse_module_h m);
static gboolean _ms_ipc_dispatch_no_instance(muse_module_h m, void *jobj);
static gpointer _ms_ipc_dispatch_worker(gpointer data);
static gboolean _ms_ipc_data_processing(int fd, muse_recv_data_head_t *header, muse_channel_info_t *ch);
static gpointer _ms_ipc_data_worker(gpointer data);

static void _ms_ipc_data_ch_init(muse_module_h m)
{
	muse_return_if_fail(m);

	m->ch[MUSE_CHANNEL_DATA].data_queue = g_queue_new();
	g_mutex_init(&m->ch[MUSE_CHANNEL_DATA].data_mutex);
	g_cond_init(&m->ch[MUSE_CHANNEL_DATA].data_cond);
}

static void _ms_ipc_data_ch_deinit(muse_module_h m)
{
	muse_return_if_fail(m);

	g_mutex_lock(&m->ch[MUSE_CHANNEL_DATA].data_mutex);
	g_queue_free(m->ch[MUSE_CHANNEL_DATA].data_queue);
	m->ch[MUSE_CHANNEL_DATA].data_queue = NULL;
	g_cond_broadcast(&m->ch[MUSE_CHANNEL_DATA].data_cond);
	g_mutex_unlock(&m->ch[MUSE_CHANNEL_DATA].data_mutex);

	g_mutex_clear(&m->ch[MUSE_CHANNEL_DATA].data_mutex);
	g_cond_clear(&m->ch[MUSE_CHANNEL_DATA].data_cond);
}

static void _ms_ipc_module_data_ch_cleanup(muse_module_h m)
{
	muse_return_if_fail(m);
	muse_return_if_fail(m->ch[MUSE_CHANNEL_DATA].thread);

	g_thread_join(m->ch[MUSE_CHANNEL_DATA].thread);
	m->ch[MUSE_CHANNEL_DATA].thread = NULL;

	_ms_ipc_data_ch_deinit(m);

	LOGD("[close] [%d] MUSE_CHANNEL_DATA", m->ch[MUSE_CHANNEL_DATA].sock_fd);
	muse_core_connection_close(m->ch[MUSE_CHANNEL_DATA].sock_fd);

}

static void _ms_ipc_module_msg_ch_cleanup(muse_module_h m)
{
	muse_return_if_fail(m);
	muse_return_if_fail(m->ch[MUSE_CHANNEL_MSG].thread);

	SECURE_LOGD("[close] [%d] MUSE_CHANNEL_MSG %p", m->ch[MUSE_CHANNEL_MSG].sock_fd, m);
	muse_core_connection_close(m->ch[MUSE_CHANNEL_MSG].sock_fd);

	SECURE_LOGD("%p thread exit", m->ch[MUSE_CHANNEL_MSG].thread);
	g_thread_unref(m->ch[MUSE_CHANNEL_MSG].thread);
	m->ch[MUSE_CHANNEL_MSG].thread = NULL;
}

static void _ms_ipc_module_cleanup(muse_module_h m)
{
	muse_return_if_fail(m);

	_ms_ipc_module_data_ch_cleanup(m);

	ms_connection_unregister(m);

	_ms_ipc_module_msg_ch_cleanup(m);

	g_mutex_clear(&m->dispatch_lock);

	memset(m, 0, sizeof(muse_module_t));

	free(m);
}

static gboolean _ms_ipc_module_instance_creation_is_allowed(int module_idx)
{
	int max_instance, created_module_instance_count;

	muse_return_val_if_fail(ms_check_module_idx(module_idx), FALSE);

	max_instance = ms_config_get_max_instance(module_idx);
	created_module_instance_count = muse_server_get_module_instance_count(module_idx);

	if (max_instance == UNLIMITED_INSTANCE || created_module_instance_count < max_instance) {
		return TRUE;
	} else {
		LOGW("The number (%d) of created module instance is over the value of max instance (%d)", created_module_instance_count, max_instance);
		return FALSE;
	}
}

static gboolean _ms_ipc_get_module_idx(muse_module_h m, void *jobj)
{
	muse_return_val_if_fail(m, FALSE);
	muse_return_val_if_fail(jobj, FALSE);

	if (!muse_core_msg_object_get_value(MSG_KEY_MODULE_INDEX, jobj, MUSE_TYPE_INT, &m->idx)) {
		LOGE("Failed to get the value of module index");
		return FALSE;
	}

	return TRUE;
}

static gboolean _ms_ipc_dispatch_create(muse_module_h m, void *jobj)
{
	int pid, dispatch_ret = MM_ERROR_NONE;

	muse_return_val_if_fail(m, FALSE);
	muse_return_val_if_fail(jobj, FALSE);

	ms_module_dispatch_lock(m);

	if (!_ms_ipc_get_module_idx(m, jobj)) {
		ms_cmd_dispatch(m, MUSE_MODULE_COMMAND_DEBUG_INFO_DUMP);
		goto out;
	}

	m->ch[MUSE_CHANNEL_MSG].dll_handle = ms_module_open(m->idx);

	if (!_ms_ipc_module_instance_creation_is_allowed(m->idx)) {
		ms_cmd_dispatch(m, MUSE_MODULE_COMMAND_RESOURCE_NOT_AVAILABLE);
		goto out;
	}

	if (muse_core_msg_object_get_value(MSG_KEY_PID, jobj, MUSE_TYPE_INT, &pid) && m->pid != pid)
		LOGW("connected pid [%d] msg [%d] is different", m->pid, pid);

	ms_connection_register(m);

	if (muse_server_is_ready())
		ms_cmd_dispatch(m, MUSE_MODULE_COMMAND_CREATE_SERVER_ACK);
	else
		LOGW("Do not send server acknowledgement because muse server is actually not ready");

	_ms_ipc_data_ch_init(m);

	LOGD("module fd: %d dll_handle: %p", m->ch[MUSE_CHANNEL_MSG].sock_fd, m->ch[MUSE_CHANNEL_MSG].dll_handle);
	dispatch_ret = ms_module_dispatch(m);

	if (dispatch_ret != MM_ERROR_NONE) {
		LOGE("create dispatch failed 0x%x, clean up module", dispatch_ret);
		goto out;
	}

	m->is_created = TRUE;

	ms_module_dispatch_unlock(m);

	return TRUE;

out:
	ms_module_dispatch_unlock(m);

	return FALSE;
}

static gboolean _ms_ipc_dispatch_destroy(muse_module_h m)
{
	int dispatch_ret = MM_ERROR_NONE;

	muse_return_val_if_fail(m, FALSE);

	ms_module_dispatch_lock(m);
	dispatch_ret = ms_module_dispatch(m);
	if (dispatch_ret != MM_ERROR_NONE)
		LOGE("destroy dispatch failed 0x%x", dispatch_ret);
	ms_module_dispatch_unlock(m);

	return FALSE;
}

static gboolean _ms_ipc_dispatch_no_instance(muse_module_h m, void *jobj)
{
	muse_return_val_if_fail(m, FALSE);
	muse_return_val_if_fail(jobj, FALSE);

	ms_module_dispatch_lock(m);

	if (!_ms_ipc_get_module_idx(m, jobj)) {
		ms_cmd_dispatch(m, MUSE_MODULE_COMMAND_DEBUG_INFO_DUMP);
		goto out;
	}

	m->ch[MUSE_CHANNEL_MSG].dll_handle = ms_module_open(m->idx);

	if (!_ms_ipc_module_instance_creation_is_allowed(m->idx)) {
		ms_cmd_dispatch(m, MUSE_MODULE_COMMAND_RESOURCE_NOT_AVAILABLE);
		goto out;
	}

	ms_connection_register(m);

	_ms_ipc_data_ch_init(m);

	ms_module_dispatch(m);

	SECURE_LOGW("[module %p] [loaded value %d]", m, ms_module_get_loaded_dllsym(m->idx));

out:
	ms_module_dispatch_unlock(m);

	return FALSE;
}

static gpointer _ms_ipc_dispatch_worker(gpointer data)
{
	int len, fd;
	int parse_len = 0;
	muse_module_h m = NULL;
	gboolean attempt_to_dispatch = TRUE;
	void *jobj = NULL;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	int tbm_fd[MUSE_NUM_FD] = { MUSE_ERR, };

	muse_return_val_if_fail(data, NULL);

	m = (muse_module_h)data;
	muse_return_val_if_fail(m, NULL);

	fd = m->ch[MUSE_CHANNEL_MSG].sock_fd;

	LOGD("Enter %d", fd);

	while (attempt_to_dispatch) {
		memset(m->recv_msg, 0x00, sizeof(m->recv_msg));

		len = muse_core_msg_recv_fd(fd, m->recv_msg, tbm_fd);
		if (len <= 0) {
			strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
			LOGE("[%s] [%d] recv : %s (%d)", ms_config_get_host_name(m->idx), fd, err_msg, errno);
			ms_cmd_dispatch(m, MUSE_MODULE_COMMAND_SHUTDOWN);
			attempt_to_dispatch = FALSE;
		}

		m->msg_offset = 0;

		while (attempt_to_dispatch && m->msg_offset < len && ms_is_server_ready()) {
			jobj = muse_core_msg_object_new(m->recv_msg + m->msg_offset, &parse_len, NULL);
			if (!jobj) {
				LOGE("jobj is null");
				attempt_to_dispatch = FALSE;
				break;
			}

			if (muse_core_msg_object_get_value(MSG_KEY_API, jobj, MUSE_TYPE_INT, &m->api)) {
				switch (m->api) {
				case API_CREATE:
					SECURE_LOGI("CREATE %p %d", m, fd);
					attempt_to_dispatch = _ms_ipc_dispatch_create(m, jobj);
					break;
				case API_DESTROY:
					SECURE_LOGI("DESTROY %p %d", m, fd);
					attempt_to_dispatch = _ms_ipc_dispatch_destroy(m);
					break;
				default:
					m->ch[MUSE_CHANNEL_MSG].tbm_fd = tbm_fd[0];
					if (m->is_created) /* handle based */
						ms_module_dispatch(m);
					else
						attempt_to_dispatch = _ms_ipc_dispatch_no_instance(m, jobj);
					break;
				}
			}
			muse_core_msg_object_free(jobj);
			jobj = NULL;
			m->msg_offset += parse_len;
			parse_len = len - parse_len;
		}
	}

	_ms_ipc_module_cleanup(m);

	LOGD("worker exit");
	g_thread_exit(NULL);

	return NULL;
}

static gboolean _ms_ipc_data_processing(int fd, muse_recv_data_head_t *header, muse_channel_info_t *ch)
{
	char *raw_data = NULL;

	muse_return_val_if_fail(ms_get_instance(), FALSE);

	if (!(fd > 0 && header && ch)) {
		LOGE("invalid param %d %p %p", fd, header, ch);
		goto _PROCESSING_FAILED;
	}

	/* check marker */
	if (header->marker != MUSE_DATA_HEAD) {
		LOGE("invalid marker 0x%x", header->marker);
		goto _PROCESSING_FAILED;
	}

	/* check data size */
	if (header->size > DATA_WORKER_QDATA_MAX_SIZE) {
		LOGE("invalid data size %d", header->size);
		goto _PROCESSING_FAILED;
	}

	/* allocation data */
	raw_data = (char *)g_try_new0(char, header->size + sizeof(muse_recv_data_head_t));
	if (!raw_data) {
		LOGE("failed to alloc data %d + %zu", header->size, sizeof(muse_recv_data_head_t));
		goto _PROCESSING_FAILED;
	}

	/* copy header */
	memcpy(raw_data, header, sizeof(muse_recv_data_head_t));

	/* receive data */
	if (!muse_core_msg_recv_len(fd, raw_data + sizeof(muse_recv_data_head_t), header->size)) {
		LOGE("receive data failed - length %d", header->size);
		goto _PROCESSING_FAILED;
	}

	/* push data */
	g_mutex_lock(&ch->data_mutex);
	g_queue_push_tail(ch->data_queue, (gpointer)raw_data);
	g_cond_signal(&ch->data_cond);
	g_mutex_unlock(&ch->data_mutex);

	return TRUE;

_PROCESSING_FAILED:

	MUSE_G_FREE(raw_data);

	ms_log_process_info(ms_get_instance()->pid);

	return FALSE;
}

static gpointer _ms_ipc_data_worker(gpointer data)
{
	char recv_buf[MUSE_MSG_LEN_MAX] = {'\0',};
	int fd;
	muse_module_h m = NULL;
	muse_channel_info_t *ch = NULL;

	muse_return_val_if_fail(data, NULL);

	m = (muse_module_h)data;
	SECURE_LOGW("module : %p", m);

	fd = m->ch[MUSE_CHANNEL_DATA].sock_fd;
	ch = &m->ch[MUSE_CHANNEL_DATA];

	/* get data */
	while (1) {
		if (!muse_core_msg_recv_len(fd, recv_buf, sizeof(muse_recv_data_head_t)))
			break;

		if (!_ms_ipc_data_processing(fd, (muse_recv_data_head_t *)recv_buf, ch)) {
			LOGE("ipc data processing failed");
			break;
		}
	}

	LOGW("Leave");

	return NULL;
}

gboolean ms_ipc_create_msg_dispatch_worker(muse_module_h m)
{
	GError *error = NULL;

	LOGD("Enter");

	muse_return_val_if_fail(ms_get_instance(), FALSE);
	muse_return_val_if_fail(ms_is_server_ready(), FALSE);

	muse_return_val_if_fail(muse_server_module_is_valid(m), FALSE);

	SECURE_LOGD("[PID %d %p] module's msg channel fd : %d", m->pid, m, m->ch[MUSE_CHANNEL_MSG].sock_fd);

	m->ch[MUSE_CHANNEL_MSG].thread = g_thread_try_new(MSG_THREAD_NAME, _ms_ipc_dispatch_worker, (gpointer)m, &error);
	if (!m->ch[MUSE_CHANNEL_MSG].thread && error) {
		LOGE("thread creation failed : %s", error->message);
		g_error_free(error);
		ms_log_process_info(ms_get_instance()->pid);
	}

	muse_return_val_if_fail(m->ch[MUSE_CHANNEL_MSG].thread, FALSE);

	LOGD("Leave");
	return TRUE;
}

gboolean ms_ipc_create_data_dispatch_worker(muse_module_h m)
{
	GError *error = NULL;

	LOGD("Enter");

	muse_return_val_if_fail(m, FALSE);
	muse_return_val_if_fail(ms_get_instance(), FALSE);
	muse_return_val_if_fail(ms_is_server_ready(), FALSE);
	muse_return_val_if_fail(muse_server_module_is_valid(m), TRUE);

	m->ch[MUSE_CHANNEL_DATA].thread = g_thread_try_new(DATA_THREAD_NAME, _ms_ipc_data_worker, (gpointer)m, &error);
	if (!m->ch[MUSE_CHANNEL_DATA].thread && error) {
		LOGE("thread creation failed : %s", error->message);
		g_error_free(error);
		ms_log_process_info(ms_get_instance()->pid);
		ms_cmd_dispatch(m, MUSE_MODULE_COMMAND_RESOURCE_NOT_AVAILABLE);
		muse_core_connection_close(m->ch[MUSE_CHANNEL_MSG].sock_fd);
	}

	muse_return_val_if_fail(m->ch[MUSE_CHANNEL_DATA].thread, FALSE);

	LOGD("Leave");
	return TRUE;
}
