/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"
#include <dlfcn.h>

#define U32BITS				0xffffffff
#define MAX_FILE_NUM		3
#define MAX_SIZE			33554432
#define WRITE_FAIL			-1
#define BUILD_ID			"BUILD_ID"
#define delimiter			"="

static gboolean _ms_log_pid_is_valid(pid_t pid);
static void _ms_log_latest_msgs(ms_log_t *log);
static void _ms_log_pid_time(ms_log_t* log, int pid);
static void _ms_log_cmdline(ms_log_t* log, int pid);
static void _ms_log_process_info(ms_log_t* log, int pid);
static void _ms_log_api_info(ms_log_t *log);
static void _ms_log_binary_info(ms_log_t *log);
static int _ms_fd_set_block(int fd);
static int _ms_log_open_work(const char *path);
static void _ms_log_create_fd(ms_log_t *log);
static void _ms_log_set_fd(ms_log_t *log);
static void _ms_log_cache_latest_msg(ms_log_t *log, int pid, char *msg);

static gboolean _ms_log_pid_is_valid(pid_t pid)
{
	char log_buf[MUSE_MSG_LEN_MAX];

	muse_return_val_if_fail(ms_is_log_enabled(), FALSE);

	snprintf(log_buf, sizeof(log_buf), "/proc/%d", (int)pid);

	if (access(log_buf, F_OK) != 0) {
		LOGD("%d process doesn't exist", (int)pid);
		return FALSE;
	}

	return TRUE;
}

static void _ms_log_latest_msgs(ms_log_t *log)
{
	int idx = 0;

	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(log);

	g_mutex_lock(&log->lock);

	if (muse_core_fd_is_valid(log->fd)) {
		for (idx = log->cur_idx; idx < MUSE_LOG_MSG_NUM; idx++)
			ms_log_write(log->latest_msgs[idx]);

		for (idx = 0; idx < log->cur_idx; idx++)
			ms_log_write(log->latest_msgs[idx]);
	}

	g_mutex_unlock(&log->lock);
}

static void _ms_log_pid_time(ms_log_t *log, int pid)
{
	char log_buf[MUSE_MSG_LEN_MAX];
	char time_buf[MUSE_LOG_MSG_LEN];

	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(log);
	muse_return_if_fail(muse_core_fd_is_valid(log->fd));
	muse_return_if_fail(_ms_log_pid_is_valid(pid));

	muse_core_get_cur_time(time_buf);
	snprintf(log_buf, sizeof(log_buf), "[PID] %d %s", pid, time_buf);

	ms_log_write(log_buf);
}

static void _ms_log_cmdline(ms_log_t *log, int pid)
{
	char log_buf[MUSE_MSG_LEN_MAX];
	FILE *fp;

	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(log);
	muse_return_if_fail(muse_core_fd_is_valid(log->fd));
	muse_return_if_fail(_ms_log_pid_is_valid(pid));

	snprintf(log_buf, sizeof(log_buf), "/proc/%d/cmdline", pid);
	fp = fopen(log_buf, "r");
	if (fp) {
		size_t size = fread(log_buf, sizeof(char), MUSE_MSG_LEN_MAX, fp);
		log_buf[size] = '\0';

		ms_log_write(log_buf);
		fclose(fp);
	}
}

static void _ms_log_process_info(ms_log_t *log, int pid)
{
	char log_buf[MUSE_MSG_LEN_MAX];
	FILE *fp;

	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(log);
	muse_return_if_fail(muse_core_fd_is_valid(log->fd));
	muse_return_if_fail(_ms_log_pid_is_valid(pid));

	snprintf(log_buf, sizeof(log_buf), "/bin/ps -Lo pcpu,pmem,tid,comm -p %d", pid);
	fp = popen(log_buf, "r");
	if (fp) {
		ms_log_write(log_buf);

		while (fgets(log_buf, MUSE_MSG_LEN_MAX, fp)) {
			if (write(log->fd, log_buf, strlen(log_buf)) == WRITE_FAIL)
				LOGE("fail to write command info to logfile");
		}

		if (pclose(fp) == -1)
			LOGE("fail to pclose");
	}
}

static void _ms_log_api_info(ms_log_t *log)
{
	Dl_info info;
	char log_buf[MUSE_MSG_LEN_MAX];
	gpointer ptr;
	int idx;

	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(log);
	muse_return_if_fail(muse_core_fd_is_valid(log->fd));

	for (idx = 0; idx < ms_config_get_host_cnt(); idx++) {
		ms_module_get_dllsym_value(idx, DISPATCHER_PTR, &ptr);
		memset(&info, 0, sizeof(info));
		if (dladdr((const void *) ptr, &info) && info.dli_sname) {
			snprintf(log_buf, sizeof(log_buf), "[The latest called api name] %s", (char *)info.dli_sname);
			ms_log_write(log_buf);
		}
	}
}

static void _ms_log_binary_info(ms_log_t *log)
{
	char log_buf[MUSE_MSG_LEN_MAX];
	FILE *fp;
	char *label;
	char *value;
	char *ptr = NULL;

	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(log);
	muse_return_if_fail(muse_core_fd_is_valid(log->fd));

	fp = fopen("/etc/tizen-release", "r");
	muse_return_if_fail(fp);

	while (fgets(log_buf, MUSE_MSG_LEN_MAX, fp)) {
		label = strtok_r(log_buf, delimiter, &ptr);
		if (label) {
			if (strncmp(label, BUILD_ID, strlen(BUILD_ID) + 1) == 0) {
				value = strtok_r(NULL, delimiter, &ptr);
				if (value)
					ms_log_write(value);
				else
					LOGE("strtok_r (value) returns null");
			}
		} else {
			LOGE("strtok_r (label) returns null");
		}
	}

	fclose(fp);
}

static int _ms_fd_set_block(int fd)
{
	int flags;
	int ret;

	muse_return_val_if_fail(ms_is_log_enabled(), MM_ERROR_NOT_SUPPORT_API);

	flags = fcntl(fd, F_GETFL);
	ret = fcntl(fd, F_SETFL, flags & (U32BITS ^ O_NONBLOCK));

	muse_core_update_fd_state(fd);

	return ret;
}

static int _ms_log_open_work(const char *path)
{
	muse_return_val_if_fail(path, MM_ERROR_INVALID_ARGUMENT);
	muse_core_remove_symlink(path);
	return open(path, O_CREAT | O_APPEND | O_WRONLY | O_NONBLOCK, 0644);
}

static void _ms_log_create_fd(ms_log_t *log)
{
	int selected_idx = 0;
	int idx;
	struct stat st;
	char file[MAX_FILE_NUM][MUSE_MSG_LEN];

	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(log);

	for (idx = 0; idx < MAX_FILE_NUM; idx++)
		snprintf(file[idx], strlen(MUSE_LOG_FILE) + 3, "%s.%d", MUSE_LOG_FILE, idx);

	for (idx = 0; idx < MAX_FILE_NUM; idx++) {
		if (access(file[idx], F_OK) == 0) { /* if 0, then there is file */
			if (stat(file[idx], &st) == MUSE_ERR) {
				LOGE("stat %s fail", file[idx]);
				continue;
			}
			log->size = st.st_size;
			if (log->size > MAX_SIZE) {
				selected_idx = (idx + 1) % MAX_FILE_NUM;
			} else {
				selected_idx = idx;
				break;
			}
		} else {
			selected_idx = idx;
			break;
		}
	}

	LOGD("filename: %s", file[selected_idx]);
	/* open log file again */
	log->fd = _ms_log_open_work(file[selected_idx]);
	if (!muse_core_fd_is_valid(log->fd))
		LOGE("couldn't open log file %d", log->fd);
	else
		LOGD("log fd : %d", log->fd);
}

static void _ms_log_set_fd(ms_log_t *log)
{
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	if (!ms_is_log_enabled())
		return;

	muse_return_if_fail(log);

	_ms_log_create_fd(log);

	muse_return_if_fail(muse_core_fd_is_valid(log->fd));

	if (fcntl(log->fd, F_SETFD, FD_CLOEXEC) < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("unable to set CLO_EXEC on log fd %d: %s", log->fd, err_msg);
	}

	(void) _ms_fd_set_block(log->fd);
}

static void _ms_log_cache_latest_msg(ms_log_t *log, int pid, char *msg)
{
	char time_buf[MUSE_LOG_MSG_LEN];
	muse_return_if_fail(msg);
	muse_return_if_fail(log);

	g_mutex_lock(&log->lock);

	muse_core_get_cur_time(time_buf);
	snprintf(log->latest_msgs[log->cur_idx], MUSE_LOG_MSG_LEN, "[P%5d] %s %s", pid, time_buf, msg);

	log->cur_idx = (log->cur_idx + 1) % MUSE_LOG_MSG_NUM;

	g_mutex_unlock(&log->lock);
}

void ms_log_init(ms_log_t *log)
{
	muse_return_if_fail(log);

	g_mutex_init(&log->lock);
	_ms_log_set_fd(log);
}

void ms_log_deinit(ms_log_t *log)
{
	LOGD("Enter");

	muse_return_if_fail(log);

	muse_return_if_fail(ms_is_log_enabled());
	g_mutex_clear(&log->lock);

	MUSE_FREE(log);

	LOGD("Leave");
}

void ms_log_save_addr_name(int pid, gpointer addr)
{
	Dl_info info;

	muse_return_if_fail(ms_get_instance());
	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(addr);

	memset(&info, 0, sizeof(info));

	if (dladdr((const void *) addr, &info) && info.dli_sname)
		_ms_log_cache_latest_msg(ms_get_instance()->log, pid, (char *)info.dli_sname);
	else
		LOGE("fail to log api name");
}

void ms_log_debug_pid_info(int pid)
{
	ms_log_t *log = NULL;

	muse_return_if_fail(ms_get_instance());
	muse_return_if_fail(ms_is_log_enabled());

	log = ms_get_instance()->log;
	muse_return_if_fail(log);

	/* pid */
	_ms_log_pid_time(log, pid);

	/* cmdline */
	_ms_log_cmdline(log, pid);

	/* process info */
	_ms_log_process_info(log, pid);

	/* api info */
	_ms_log_api_info(log);

	/* binary version */
	_ms_log_binary_info(log);

	/* latest api */
	_ms_log_latest_msgs(log);
}

void ms_log_write(char *buf)
{
	ms_log_t *log = NULL;

	muse_return_if_fail(ms_get_instance());

	log = ms_get_instance()->log;

	muse_return_if_fail(ms_is_log_enabled());
	muse_return_if_fail(log);
	muse_return_if_fail(muse_core_fd_is_valid(log->fd));
	muse_return_if_fail(buf);

	if (write(log->fd, buf, strlen(buf)) == WRITE_FAIL)
		LOGE("There was an error writing module pid to logfile");
	else if (write(log->fd, "\n", 1) == WRITE_FAIL)
		LOGE("fail to write new line character");
}

