/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"
#ifdef MUSE_USE_RM_READY
#include "mm_resource_manager.h"
#endif

static GMutex dllsym_table_lock;

static gboolean _ms_module_dispatch_timeout_callback(gpointer data);
static gboolean _ms_module_enable_dispatch_timeout_callback(int cmd, muse_module_h m);
static void _ms_module_set_loaded_dllsym(int idx, GModule *dllsym, gboolean value);

static gboolean _ms_module_dispatch_timeout_callback(gpointer data)
{
	muse_module_h m = (muse_module_h)data;
	char err_msg[MUSE_MSG_LEN_MAX];
	int idx, api, timeout;

	muse_return_val_if_fail(ms_get_instance(), FALSE);

	LOGE("Restart muse server");

	if (m) {
		idx = m->idx;
		api = m->api;
		timeout = ms_module_get_timeout(idx, api);
		snprintf(err_msg, sizeof(err_msg), "[DEADLOCK] %s (pid %d)'s dispatcher (%d) does not return value within (%d) second",
					ms_config_get_host_name(idx), m->pid, api, timeout);

		LOGE("%s", err_msg);

		ms_log_process_info(ms_get_instance()->pid);
		if (ms_is_log_enabled())
			ms_log_write(err_msg);
	} else {
		LOGE("invalid pointer for module");
	}

	ms_respawn(SIGABRT);

	return FALSE;
}

static gboolean _ms_module_enable_dispatch_timeout_callback(int cmd, muse_module_h m)
{
	muse_return_val_if_fail(m, FALSE);

	if (cmd == ms_config_get_disable_watchdog_api(m->idx))
		return FALSE;

	return TRUE;
}

static void _ms_module_set_loaded_dllsym(int idx, GModule *dllsym, gboolean value)
{
	ms_module_t *module = ms_get_module_instance(idx);

	muse_return_if_fail(module);

	module->loaded = value;
	module->dllsym = dllsym;
	SECURE_LOGD("dll_handle: %p", module->dllsym);
}

GModule *ms_module_open(int idx)
{
	GModule *dllsym = NULL;
	ms_module_t *module = ms_get_module_instance(idx);

	muse_return_val_if_fail(module, NULL);

	g_mutex_lock(&module->lock);

	if (!ms_module_get_loaded_dllsym(idx)) {
		dllsym = g_module_open(ms_config_get_path(idx), G_MODULE_BIND_LAZY);

		if (!dllsym)
			LOGE("%s", g_module_error());
		else
			_ms_module_set_loaded_dllsym(idx, dllsym, true);
	} else {
		dllsym = ms_module_get_dllsym(idx);
		SECURE_LOGW("already module is opened: %p", dllsym);
	}

	g_mutex_unlock(&module->lock);

	return dllsym;
}

void ms_module_dispatch_lock(muse_module_h m)
{
	muse_return_if_fail(m);
	g_mutex_lock(&m->dispatch_lock);
}

void ms_module_dispatch_unlock(muse_module_h m)
{
	muse_return_if_fail(m);
	g_mutex_unlock(&m->dispatch_lock);
}

int ms_module_dispatch(muse_module_h m)
{
	int ret = MUSE_ERR;
	guint id = 0;
	gboolean rm_success = FALSE;
	gboolean enable_dispatch_timeout_callback = FALSE;
	muse_module_dispatchfunc *dispatcher = NULL;
	ms_module_t *module = NULL;
	int api = 0;

	muse_return_val_if_fail(m, MUSE_ERR);
	api = m->api;

	module = ms_get_module_instance(m->idx);
	muse_return_val_if_fail(module, MUSE_ERR);
	muse_return_val_if_fail(module->dllsym, MUSE_ERR);
	muse_return_val_if_fail(api < API_MAX, MUSE_ERR);

#ifdef MUSE_USE_RM_READY
	if (access(MM_RESOURCE_MANAGER_READY, F_OK) != 0) {
		LOGE("We skip dispatching (%d) due to resource-manager unavailable", api);
		ms_cmd_dispatch(m, MUSE_MODULE_COMMAND_RESOURCE_MANAGER_SHUTDOWN);
		return MUSE_ERR;
	}
#endif

	g_mutex_lock(&module->lock);
	g_module_symbol(module->dllsym, DISPATCHER, (gpointer *)&dispatcher);
	g_mutex_unlock(&module->lock);

	if (dispatcher && dispatcher[api]) {
		if (ms_is_log_enabled()) {
			ms_module_set_dllsym_value(m->idx, DISPATCHER_PTR, (gpointer)dispatcher[api]);
			ms_log_save_addr_name(m->pid, (gpointer)dispatcher[api]);
		}

		enable_dispatch_timeout_callback = _ms_module_enable_dispatch_timeout_callback(api, m);

		if (enable_dispatch_timeout_callback)
			id = g_timeout_add_seconds((guint)module->disp_timeout[api], (GSourceFunc)_ms_module_dispatch_timeout_callback, (gpointer)m);

		ret = dispatcher[api](m);

		if (enable_dispatch_timeout_callback) {
			rm_success = g_source_remove(id);
			if (!rm_success)
				LOGE("[%s] [%d] Failed to remove %d", ms_config_get_host_name(m->idx), api, id);
		}
	} else {
		LOGE("[%s] [%d] error - dispatcher", ms_config_get_host_name(m->idx), api);
		ret = MUSE_ERR;
	}

	return ret;
}

gboolean ms_module_close(muse_module_h m)
{
	ms_module_t *module = NULL;

	muse_return_val_if_fail(m, FALSE);

	module = ms_get_module_instance(m->idx);
	muse_return_val_if_fail(module, FALSE);

	g_mutex_lock(&module->lock);

	LOGD("Closing module %s", g_module_name(m->ch[MUSE_CHANNEL_MSG].dll_handle));
	if (!g_module_close(m->ch[MUSE_CHANNEL_MSG].dll_handle)) {
		LOGE("Couldn't close dll_handle %s", g_module_error());
		g_mutex_unlock(&module->lock);
		return FALSE;
	}

	_ms_module_set_loaded_dllsym(m->idx, NULL, false);

	g_mutex_unlock(&module->lock);

	return TRUE;
}

GModule *ms_module_get_dllsym(int idx)
{
	ms_module_t *module = ms_get_module_instance(idx);
	muse_return_val_if_fail(module, NULL);

	return module->dllsym;
}

gboolean ms_module_get_loaded_dllsym(int idx)
{
	ms_module_t *module = ms_get_module_instance(idx);
	muse_return_val_if_fail(module, false);

	return module->loaded;
}

void ms_module_set_dllsym_value(int idx, const char *name, gpointer value)
{
	ms_module_t *module = ms_get_module_instance(idx);

	muse_return_if_fail(module);
	muse_return_if_fail(name);

	g_mutex_lock(&dllsym_table_lock);
	g_hash_table_insert(module->dllsym_table, g_strdup(name), value);
	g_mutex_unlock(&dllsym_table_lock);
}

int ms_module_get_dllsym_value(int idx, const char *name, gpointer *value)
{
	ms_module_t *module = ms_get_module_instance(idx);

	muse_return_val_if_fail(module, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(name, MM_ERROR_INVALID_ARGUMENT);

	if (value) {
		*value = g_hash_table_lookup(module->dllsym_table, name);
		return MM_ERROR_NONE;
	} else {
		return MM_ERROR_INVALID_ARGUMENT;
	}
}

void ms_module_set_timeout(int idx, int api, int timeout)
{
	ms_module_t *module = ms_get_module_instance(idx);

	muse_return_if_fail(module);
	muse_return_if_fail(timeout >= ms_config_get_min_timeout() && timeout <= ms_config_get_max_timeout());

	g_mutex_lock(&module->lock);

	module->disp_timeout[api] = timeout;

	g_mutex_unlock(&module->lock);
}

int ms_module_get_timeout(int idx, int api)
{
	ms_module_t *module = ms_get_module_instance(idx);
	muse_return_val_if_fail(module, MUSE_ERR);

	return module->disp_timeout[api];
}

void ms_module_deinit(ms_module_t *module)
{
	muse_return_if_fail(module);

	g_mutex_clear(&module->lock);
	g_module_close(module->dllsym);
	g_hash_table_destroy(module->dllsym_table);

	free(module);
}

void ms_module_init(ms_module_t *module)
{
	int disp_idx = 0;

	LOGD("Enter");

	muse_return_if_fail(module);

	g_mutex_init(&module->lock);
	module->loaded = FALSE;
	module->dllsym_table = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
	muse_return_if_fail(module->dllsym_table);
	for (disp_idx = 0; disp_idx < API_MAX; disp_idx++)
		ms_module_set_timeout(module->idx, disp_idx, ms_config_get_timeout(module->idx));

	LOGD("Leave");
}

