/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"
#include <sys/file.h>
#include <sys/syscall.h>
#include <gst/gst.h>
#include <syslog.h>

#ifdef MUSE_REGISTER_VIP
#include <proc_stat.h>
#endif

#ifdef MUSE_USE_LWIPC
#include <lwipc.h>
#define MUSE_LWIPC_WAIT_TIME			1000
#endif

static const char *channel_name[MUSE_CHANNEL_MAX] = {
	"msg",
	"data"
};

const char *UDS_files[MUSE_CHANNEL_MAX] = {MUSE_SOCK_FILE0, MUSE_SOCK_FILE1};

static muse_server_h muse_server;

static const char *module_cmd[MUSE_MODULE_COMMAND_MAX] = {
	"initialize",
	"shutdown",
	"debug_info_dump",
	"create_server_ack",
	"resource_not_available",
	"external_storage_state_changed",
	"create_caution",
};

static bool _ms_attach(int fd, muse_module_callback connection_handler, gpointer module_idx);
static void _ms_create_new_server_from_fd(int fd[], int type);
static int _ms_new(muse_channel_e channel);
static int _ms_get_pid(int fd);
static void _ms_get_module_addr(int fd, intptr_t *module_addr);
static void _ms_check_idle_state(void);
static void _ms_check_connection_event(void);
gpointer _ms_diag_thread(gpointer data);
static void _ms_lock_state(void);
static void _ms_unlock_state(void);
static gboolean _ms_connection_handler(GIOChannel *source, GIOCondition condition, gpointer data);
#ifdef MUSE_USE_LWIPC
static void _ms_wait_event(void);
static gboolean _ms_idle_cb(gpointer user_data);

static void _ms_wait_event(void)
{
	const char *lw_event_list[] = { "/run/.wm_ready", "/tmp/avoc_ready" };
	unsigned int count = sizeof(lw_event_list) / sizeof(char *);

	if (LwipcWaitMultiEvents(lw_event_list, count, true, MUSE_LWIPC_WAIT_TIME, NULL, 0) != 0)
		LOGE("Fail to receive Multiple Events");
}
#endif

static bool _ms_attach(int fd, muse_module_callback connection_handler, gpointer module_idx)
{
	GIOChannel *channel = NULL;
	GSource *src = NULL;

	LOGI("Enter");

	muse_return_val_if_fail(muse_server, false);
	muse_return_val_if_fail(muse_core_fd_is_valid(fd), false);

	channel = g_io_channel_unix_new(fd);
	muse_return_val_if_fail(channel, false);

	src = g_io_create_watch(channel, G_IO_IN);
	if (!src) {
		LOGE("g_io_create_watch() is failed");
		g_io_channel_unref(channel);
		return false;
	}

	g_source_set_callback(src, (GSourceFunc)connection_handler, module_idx, NULL);

	if (g_source_attach(src, g_main_loop_get_context(muse_server->main_loop)) == 0) {
		LOGE("g_source_attach() is failed");
		g_io_channel_unref(channel);
		return false;
	}

	g_source_unref(src);

	g_io_channel_unref(channel);

	LOGI("Leave");

	return true;
}

static void _ms_create_new_server_from_fd(int fd[], int type)
{
	int i;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	LOGD("Enter");

	muse_return_if_fail(muse_server);

	muse_server->msg_fd = fd[MUSE_CHANNEL_MSG];
	muse_server->data_fd = fd[MUSE_CHANNEL_DATA];
	muse_server->type = type;

	gettimeofday(&muse_server->tv_s, NULL);

	for (i = 0; i < MUSE_CHANNEL_MAX; i++) {
		if (!_ms_attach(fd[i], _ms_connection_handler, (gpointer)(intptr_t) i)) {
			snprintf(err_msg, sizeof(err_msg), "Fail to attach server fd %d", fd[i]);

			LOGE("%s", err_msg);
			ms_respawn(SIGABRT);
			return;
		}
	}

	LOGD("Leave");
}

static int _ms_new(muse_channel_e channel)
{
	int fd, errsv;
	struct sockaddr_un addr_un;
	socklen_t address_len;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	muse_return_val_if_fail(channel < MUSE_CHANNEL_MAX, MM_ERROR_INVALID_ARGUMENT);

	unlink(UDS_files[channel]);

	/* Create Socket */
	fd = socket(AF_UNIX, SOCK_STREAM, 0); /* Unix Domain Socket */
	if (!muse_core_fd_is_valid(fd)) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("socket failed sock (%d) : %s", errno, err_msg);
		return MUSE_ERR;
	}

	LOGD("muse server fd : %d", fd);

	memset(&addr_un, 0, sizeof(addr_un));
	addr_un.sun_family = AF_UNIX;
	strncpy(addr_un.sun_path, UDS_files[channel], sizeof(addr_un.sun_path) - 1);
	address_len = sizeof(addr_un);

	/* Bind to filename */
	if (bind(fd, (struct sockaddr *)&addr_un, address_len) < 0) {
		errsv = errno;
		strerror_r(errsv, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] socket bind failed (%d) %s", fd, errsv, err_msg);
		muse_core_log_file_list(UDS_files[channel]);
		ms_log_user_group_info();
		if (errsv == EADDRINUSE)
			unlink(addr_un.sun_path);
		close(fd);
		return MUSE_ERR;
	}

	/* Setup listen queue */
	if (listen(fd, 5) == MUSE_ERR) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("[%d] listen failed (%d) %s", fd, errno, err_msg);
		muse_core_log_file_list(UDS_files[channel]);
		ms_log_user_group_info();
		close(fd);
		return MUSE_ERR;
	}

	if (muse_core_set_nonblocking(fd, false) < 0) /* blocking */
		LOGE("failed to set server socket to blocking");

	return fd;
}

static int _ms_get_pid(int fd)
{
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	struct ucred credentials;
	socklen_t length;

	length = sizeof(struct ucred);
	if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &credentials, &length) < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("failed to get the value of credential type %s", err_msg);
		return MUSE_ERR;
	}

	muse_core_update_fd_state(fd);

	return credentials.pid;
}

static void _ms_get_module_addr(int fd, intptr_t *module_addr)
{
	void *jobj;
	int try_count = 0;
	bool ret = true;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	char recv_buf[MUSE_MSG_LEN_MAX] = {'\0',};

	do {
		if (muse_core_msg_recv_fd(fd, recv_buf, NULL) <= 0) {
			strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
			LOGE("failed to receive message for module %s", err_msg);
			return;
		}

		jobj = muse_core_msg_object_new(recv_buf, NULL, NULL);
		if (jobj) {
			ret = muse_core_msg_object_get_value(MSG_KEY_MODULE_ADDR, jobj, MUSE_TYPE_POINTER, module_addr);
			muse_core_msg_object_free(jobj);
			if (!ret)
				LOGE("[%d] Error - module_addr %s", try_count, recv_buf);
			else
				break;
		}
	} while (++try_count < MS_RECV_TRY_COUNT_MAX);
}

static gboolean _ms_connection_handler(GIOChannel *source, GIOCondition condition, gpointer data)
{
	int server_sockfd, client_sockfd, pid, idx, len;
	socklen_t client_len;
	struct sockaddr_un client_address;
	muse_channel_e channel = (muse_channel_e)data;
	muse_module_h m = NULL;
	muse_module_h peeked_m = NULL;
	muse_module_h candidate_m = NULL;
	intptr_t module_addr = 0;
	GQueue *instance_queue = NULL;
	ms_connection_t *connection = NULL;

	LOGI("Enter");

	muse_return_val_if_fail(muse_server, FALSE);

	connection = muse_server->connection;
	muse_return_val_if_fail(connection, FALSE);

	_ms_lock_state();

	if (!ms_is_server_ready()) {
		LOGW("Now mused state is not ready...");
		_ms_unlock_state();
		return FALSE;
	}

	server_sockfd = g_io_channel_unix_get_fd(source);
	if (!muse_core_fd_is_valid(server_sockfd)) {
		LOGE("Critical Error : server %d is invalid", server_sockfd);
		_ms_unlock_state();
		muse_core_dump_fd_state(server_sockfd);
		ms_respawn(SIGABRT);
		return FALSE;
	}

	client_len = sizeof(client_address);

	LOGI("[%d] Try to accept...", server_sockfd);
	client_sockfd = accept(server_sockfd, (struct sockaddr *)&client_address, &client_len);
	if (!muse_core_fd_is_valid(client_sockfd)) {
		LOGE("Critical Error : accept %d is invalid", client_sockfd);
		muse_core_dump_fd_state(client_sockfd);
		close(server_sockfd);
		_ms_unlock_state();
		ms_respawn(SIGABRT);
		return FALSE;
	}

	LOGI("server : %d client [%s channel] : %d", server_sockfd, channel_name[channel], client_sockfd);

	pid = _ms_get_pid(client_sockfd);

	if (channel == MUSE_CHANNEL_MSG) {
		m = calloc(1, sizeof(muse_module_t));
		if (!m) {
			LOGE("module allocation failure");
			goto out;
		}
		m->ch[MUSE_CHANNEL_MSG].sock_fd = client_sockfd;
		m->pid = pid;
		g_mutex_init(&m->dispatch_lock);

		ms_ipc_create_msg_dispatch_worker(m);

	} else {
		_ms_get_module_addr(client_sockfd, &module_addr);

		ms_connection_lock(connection);

		instance_queue = connection->instance_queue;
		len = g_queue_get_length(instance_queue);

		m = (muse_module_h)module_addr;

		for (idx = 0; idx < len; idx++) {
			peeked_m = (muse_module_h)g_queue_peek_nth(instance_queue, idx);
			if (!peeked_m) {
				LOGW("[%d] Make sure if the queue length is changed (%d = %d), which means that it was destroyed somewhere",
					idx, len, g_queue_get_length(instance_queue));
				continue;
			}

			if (peeked_m->pid != pid)
				continue;

			if (!m) {
				if (candidate_m) {
					LOGE("muse-server can't support the error case which threre are several modules now");
					ms_connection_unlock(connection);
					goto out;
				}

				if (!muse_core_fd_is_valid(peeked_m->ch[MUSE_CHANNEL_DATA].sock_fd))
					candidate_m = peeked_m;
				else
					SECURE_LOGW("already paired module %p", peeked_m);

				continue;
			}

			if (m != peeked_m)
				continue;

			if (muse_core_fd_is_valid(m->ch[MUSE_CHANNEL_DATA].sock_fd)) {
				SECURE_LOGE("[%d] %s pid %d %p you had better check if instance destroy completed properly",
					client_sockfd, ms_config_get_host_name(m->idx), pid, m);
				ms_connection_unlock(connection);
				goto out;
			}

			m->ch[MUSE_CHANNEL_DATA].sock_fd = client_sockfd;
			SECURE_LOGI("%s (pid %d) module : %p module addr from client : %p",
				ms_config_get_host_name(m->idx), pid, m, (void *)module_addr);
			break;
		}

		if (candidate_m) {
			m = candidate_m;
			m->ch[MUSE_CHANNEL_DATA].sock_fd = client_sockfd;
			SECURE_LOGW("[%d] %s pid %d %p restore module address at the only one null data channel",
				client_sockfd, ms_config_get_host_name(m->idx), pid, m);
		}

		ms_connection_unlock(connection);

		ms_ipc_create_data_dispatch_worker(m);
	}

	_ms_unlock_state();

	LOGI("Leave");

	return TRUE;
out:
	close(server_sockfd);
	close(client_sockfd);

	if (m) {
		if (channel == MUSE_CHANNEL_MSG)
			free(m);
		else
			muse_core_connection_close(m->ch[MUSE_CHANNEL_MSG].sock_fd);
	}


	_ms_unlock_state();

	LOGE("FALSE");

	return FALSE;
}

static void _ms_check_idle_state(void)
{
	struct timeval tv_c, tv_r;
	int instance_number, timeout;

	muse_return_if_fail(muse_server);
	muse_return_if_fail(muse_server->state == MUSE_SERVER_STATE_READY);

	gettimeofday(&tv_c, NULL);
	timersub(&tv_c, &muse_server->tv_s, &tv_r);

	instance_number = g_queue_get_length(muse_server->connection->instance_queue);
	timeout = (int)tv_r.tv_sec;

	LOGD("timeout %d", timeout);

	if (instance_number == 0 && timeout >= ms_config_get_max_idle_time()) {
		LOGE("Timeout exit !!! [Idle time] %d sec", timeout);
		ms_remove_ready_file();
		exit(EXIT_SUCCESS);
	}
}

static void _ms_check_connection_event(void)
{
	ms_connection_state_e state = MUSE_CONNECTION_STATE_INVALID;

	muse_return_if_fail(muse_server);

	if (ms_connection_get_state(&state)) {
		gettimeofday(&muse_server->tv_s, NULL);
		if (state == MUSE_CONNECTION_STATE_CONNECTED) {
			/* will be updated about connection at the next patch */
		} else if (state == MUSE_CONNECTION_STATE_DISCONNECTED) {
			LOGD("Diagnostic thread checks the memory of idle");
			ms_check_memory(muse_server->pid);
		}
	}
}

gpointer _ms_diag_thread(gpointer data)
{
	muse_return_val_if_fail(muse_server, NULL);

	while (1) {
		if (muse_server->conf->is_on_demand)
			_ms_check_idle_state();

		_ms_check_connection_event();
	}

	return NULL;
}

static void _ms_lock_state(void)
{
	muse_return_if_fail(muse_server);
	g_mutex_lock(&muse_server->state_lock);
}

static void _ms_unlock_state(void)
{
	muse_return_if_fail(muse_server);
	g_mutex_unlock(&muse_server->state_lock);
}

static gboolean _ms_idle_cb(gpointer user_data)
{
	if (!ms_create_ready_file())
		LOGE("%s file creation is failed", MUSE_SERVER_READY);

	return G_SOURCE_REMOVE;
}

static void _ms_init(void)
{
	int idx;

	LOGD("Enter");

	muse_server = calloc(1, sizeof(muse_server_t));
	muse_return_if_fail(muse_server);

	muse_server->system = calloc(1, sizeof(ms_system_t));
	muse_return_if_fail(muse_server->system);
	ms_system_init(muse_server->system);

	muse_server->conf = calloc(1, sizeof(ms_config_t));
	muse_return_if_fail(muse_server->conf);
	ms_config_init(muse_server->conf);

	muse_server->log = calloc(1, sizeof(ms_log_t));
	muse_return_if_fail(muse_server->log);
	ms_log_init(muse_server->log);

	muse_server->security = calloc(1, sizeof(ms_security_t));
	muse_return_if_fail(muse_server->security);
	ms_security_init(muse_server->security);

	for (idx = 0; idx < muse_server->conf->host_cnt; idx++) {
		muse_server->module[idx] = calloc(1, sizeof(ms_module_t));
		muse_return_if_fail(muse_server->module[idx]);
		muse_server->module[idx]->idx = idx;
		ms_module_init(muse_server->module[idx]);
	}

#ifdef MUSE_USE_WATCHDOG
	muse_server->watchdog = calloc(1, sizeof(ms_watchdog_t));
	muse_return_if_fail(muse_server->watchdog);
	if (ms_watchdog_init(muse_server->watchdog) != MM_ERROR_NONE)
		LOGE("Fail to initialize server watchdog");
#endif

	muse_server->connection = calloc(1, sizeof(ms_connection_t));
	muse_return_if_fail(muse_server->connection);
	ms_connection_init(muse_server->connection);

	ms_signal_init();

	g_mutex_init(&muse_server->state_lock);

	muse_core_create_fd_table();

	muse_server->main_loop = g_main_loop_new(NULL, FALSE);
	muse_return_if_fail(muse_server->main_loop);

	LOGD("Leave");
}

int ms_open_lockfile(void)
{
	int fd, already_running;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	char *lockfile = NULL;

	muse_return_val_if_fail(muse_server, MUSE_ERR);

	lockfile = ms_config_get_lockfile();
	muse_return_val_if_fail(lockfile, MUSE_ERR);

	muse_core_remove_symlink((const char *)lockfile);
	fd = open(lockfile, O_RDONLY);
	if (fd == -1 && errno != ENOENT) {
		/* Cannot open file even though file exists. */
		snprintf(err_msg, sizeof(err_msg), "datserver: Cannot open lock file %s", lockfile);
		LOGE("open failed : %s", err_msg);
		return MUSE_ERR;
	} else if (fd != -1) {
		already_running = flock(fd, LOCK_EX | LOCK_NB) == -1;
		close(fd);
		if (already_running) {
			LOGE("File already locked. There's already a server running");
			return MUSE_ERR;
		}
	}

	/* Lock file does not exist, or is not locked. Create a new lockfile and lock it. */
	fd = open(lockfile, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		LOGE("dataserver: Cannot create lock file");
		return MUSE_ERR;
	}

	if (flock(fd, LOCK_EX | LOCK_NB) != 0) {
		LOGE("Can't lock the lock file \"%s\". " "Is another instance running?", lockfile);
		close(fd);
		return MUSE_ERR;
	}

	close(fd);

	return MM_ERROR_NONE;
}

void ms_diag_thread_create(void)
{
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	GError *error = NULL;

	muse_return_if_fail(muse_server);

	muse_server->diag_thread = g_thread_try_new("muse_server_diag_thread", _ms_diag_thread, muse_server->main_loop, &error);
	if (!muse_server->diag_thread && error) {
		snprintf(err_msg, sizeof(err_msg), "diag_thread creation failed : %s", error->message);
		LOGE("%s", err_msg);
		g_error_free(error);
		ms_log_process_info(muse_server->pid);
	}
}

void ms_diag_thread_destroy(void)
{
	muse_return_if_fail(muse_server);
	muse_return_if_fail(muse_server->diag_thread);

	LOGD("Enter");

	g_thread_join(muse_server->diag_thread);
	muse_server->diag_thread = NULL;

	LOGD("Leave");
}

void ms_setup_syslog(void)
{
	int flags = LOG_CONS|LOG_NDELAY|LOG_PID;
	if (isatty(STDOUT_FILENO))
		flags |= LOG_PERROR;

	openlog("mused", flags, LOG_DAEMON);
	LOGD("openlog - mused");
}

void ms_fork(int *notify_fd)
{
	pid_t pid;
	int fds[2];
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	char msg[MUSE_MSG_LEN_MAX] = {'\0',};

	if (pipe(fds) == MUSE_ERR) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("Failed to create pipe to get child status: %s", err_msg);
		exit(EXIT_FAILURE);
	}

	if ((pid = fork()) < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("Error: fork() failed: %s", err_msg);
		exit(EXIT_FAILURE);
	} else if (pid != 0) {
		close(fds[1]);
		/* Read in a string from the pipe */
		if (read(fds[0], msg, sizeof(msg)) <= 0) {
			LOGE("Failed to read from a file descriptor [%d]", fds[0]);
			close(fds[0]);
			return;
		}
		close(fds[0]);

		/* Parent process closes up output side of pipe */
		if (!strcmp(msg, MSG_DONE)) {
			LOGI("Successfully daemonized");
			exit(EXIT_SUCCESS);
		} else {
			LOGE("Daemonizing failed after fork");
			exit(EXIT_FAILURE);
		}
	} else if (pid == 0) {
		/* Child process closes up input side of pipe */
		close(fds[0]);
		*notify_fd = fds[1];
	}
}

pid_t ms_daemonize(int *notify_fd)
{
	pid_t pid;
	int fd, result;

	muse_return_val_if_fail(notify_fd, MUSE_ERR);

	ms_fork(notify_fd);

	if ((pid = setsid()) < 0) {
		LOGE("create new session");
		exit(EXIT_FAILURE);
	}

	/* change the file mode mask */
	umask(0);

	result = chdir("/");
	LOGD("result = %d sid: %d pgid: %d pid: %d ppid: %d", result, (int)getsid(0), (int)getpgid(0), (int)pid, (int)getppid());

	/* redirect fds to /dev/null */
	fd = open("/dev/null", O_RDWR);
	if (!muse_core_fd_is_valid(fd)) {
		LOGE("Critical Error : %d is invalid", fd);
		exit(EXIT_SUCCESS);
	}

	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	dup2(fd, STDIN_FILENO);
	dup2(fd, STDOUT_FILENO);
	dup2(fd, STDERR_FILENO);

	close(fd);

	return pid;
}

void ms_daemonize_complete(int notify_fd)
{
	LOGW("Enter");

	muse_return_if_fail(muse_core_fd_is_valid(notify_fd));

#ifdef MUSE_REGISTER_VIP
	proc_stat_set_vip_process();
#endif

	write(notify_fd, MSG_DONE, strlen(MSG_DONE) + 1);
	LOGI("[%d] Notify parent process that child initialization is done", notify_fd);
	close(notify_fd);

	LOGW("Leave");
}

void ms_gst_init(char **cmd)
{
	gint *argc = NULL;
	gchar **argv = NULL;
	GError *err = NULL;
	gboolean ret = FALSE;
	int gst_param_cnt;

#ifdef MUSE_TTRACE_LOG
	trace_begin("MUSE:gst_init");
#endif

	gst_param_cnt = ms_config_get_gst_param_cnt();

	argc = malloc(sizeof(gint));
	muse_return_if_fail(argc);

	/* add gst_param */
	argv = malloc(sizeof(gchar *) * (gst_param_cnt + 1));
	if (!argv) {
		LOGE("argv memory leak allocatoin failed");
		free(argc);
		return;
	}

	memset(argv, 0, sizeof(gchar *) * (gst_param_cnt + 1));

	*argc = 0;
	argv[*argc] = (gchar *)cmd[0];
	(*argc)++;
	for (; (*argc) <= gst_param_cnt; (*argc)++)
		argv[*argc] = ms_config_get_gst_param_str((*argc) - 1);

	/* initializing gstreamer */
	ret = gst_init_check(argc, &argv, &err);
	if (!ret) {
		LOGE("Could not initialize GStreamer: %s ", err ? err->message : "unknown error occurred");
		if (err)
			g_error_free(err);
	}

	LOGI("gst_init_check is completed");

	/* release */
	free(argv);
	free(argc);

	LOGI("complete to initialize gstreamer");

#ifdef MUSE_TTRACE_LOG
	trace_end();
#endif
}

int ms_pidfile_create(const char *path, pid_t pid)
{
	int fd;
	struct flock lock;
	char pid_buf[MUSE_MSG_LEN] = {'\0',};
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	muse_return_val_if_fail(path, MM_ERROR_INVALID_ARGUMENT);
	muse_core_remove_symlink(path);
	fd = open(path, O_WRONLY | O_CREAT, (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));

	if (!muse_core_fd_is_valid(fd)) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("Fail to open pidfile [%s] : %s", path, err_msg);
		return MM_ERROR_FILE_NOT_FOUND;
	}

	lock.l_type = F_WRLCK;
	lock.l_start = 0;
	lock.l_whence = SEEK_SET;
	lock.l_len = 0;

	if (fcntl(fd, F_SETLK, &lock) < 0) {
		if (errno != EACCES && errno != EAGAIN) {
			strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
			LOGE("Fail to lock pidfile [%s] : %s", path, err_msg);
		} else {
			LOGE("process is already running");
		}
		close(fd);
		return MM_ERROR_FILE_INTERNAL;
	}

	if (ftruncate(fd, 0) < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("Fail to truncate pidfile [%s] : %s", path, err_msg);
		close(fd);
		return MM_ERROR_FILE_INTERNAL;
	}

	memset(pid_buf, 0, sizeof(pid_buf));
	snprintf(pid_buf, sizeof(pid_buf), "%u", pid);

	if (write(fd, pid_buf, strlen(pid_buf)) != (int)strlen(pid_buf)) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("Fail to write pid to pidfile [%s] : %s", path, err_msg);
		close(fd);
		return MM_ERROR_FILE_WRITE;
	}

	close(fd);
	return MM_ERROR_NONE;
}

void ms_init(char **argv)
{
	int idx;
	pid_t pid;
	int notify_fd;
	muse_module_cmd_dispatchfunc *cmd_dispatcher = NULL;

	LOGD("Enter");

	ms_setup_syslog();

	pid = ms_daemonize(&notify_fd);

	if (ms_pidfile_create(MUSE_DEFAULT_PIDFILE, pid) != MM_ERROR_NONE)
		exit(EXIT_FAILURE);
	else
		LOGD("MUSE_DEFAULT_PIDFILE(%s) file was created", MUSE_DEFAULT_PIDFILE);

	_ms_init();

	muse_return_if_fail(ms_get_instance());
	ms_get_instance()->pid = pid;

	muse_return_if_fail(ms_open_lockfile() == MM_ERROR_NONE);

	ms_new();

	ms_daemonize_complete(notify_fd);

#ifdef MUSE_TTRACE_LOG
	trace_end();
#endif

#ifdef MUSE_TTRACE_LOG
	trace_begin("MUSE:preloading module");
#endif
	for (idx = 0; idx < ms_config_get_host_cnt(); idx++) {
		if (0 == strncmp(ms_config_get_preloaded_value(idx), "yes", strlen("yes") + 1)) {
			g_module_symbol(ms_module_open(idx), CMD_DISPATCHER, (gpointer *)&cmd_dispatcher);
			if (cmd_dispatcher && cmd_dispatcher[MUSE_MODULE_COMMAND_INITIALIZE])
				cmd_dispatcher[MUSE_MODULE_COMMAND_INITIALIZE](NULL);
		}
	}
#ifdef MUSE_TTRACE_LOG
	trace_end();
#endif

	ms_system_subscribe_external_event(ms_get_instance()->system);

	ms_diag_thread_create();

#ifdef MUSE_USE_WATCHDOG
	if (!ms_watchdog_attach(ms_get_instance()->watchdog)) {
		LOGE("watchdog thread failed");
		ms_log_process_info(pid);
		return;
	}
#endif

	ms_gst_init(argv);

	LOGD("Leave");
}

muse_server_h ms_get_instance(void)
{
	return muse_server;
}

gboolean ms_check_module_idx(int idx)
{
	int module_cnt = ms_config_get_host_cnt();

	if (idx < 0 || idx >= module_cnt) {
		LOGE("%d error - the number of modules is %d", idx, module_cnt);
		return FALSE;
	}

	return TRUE;
}

ms_module_t *ms_get_module_instance(int idx)
{
	muse_return_val_if_fail(ms_check_module_idx(idx), NULL);

	return muse_server->module[idx];
}

int ms_deinit(void)
{
	int retval = MUSE_ERR;
	int idx;

	LOGD("Enter");

	muse_return_val_if_fail(muse_server, retval);
	muse_return_val_if_fail(muse_server->conf, retval);
	muse_return_val_if_fail(muse_server->connection, retval);
	muse_return_val_if_fail(muse_server->log, retval);
	muse_return_val_if_fail(muse_server->security, retval);
	muse_return_val_if_fail(muse_server->watchdog, retval);
	muse_return_val_if_fail(muse_server->diag_thread, retval);

	ms_recursive_rmdir(MUSE_DATA_ROOT_PATH);

	ms_set_state(MUSE_SERVER_STATE_IDLE);

	ms_diag_thread_destroy();

#ifdef MUSE_USE_WATCHDOG
	ms_watchdog_detach(muse_server->watchdog);

	if (ms_watchdog_deinit(muse_server->watchdog) == MM_ERROR_NONE)
		free(muse_server->watchdog);
	else
		LOGE("Fail to deinitialize server watchdog");
#endif

#ifdef MUSE_USE_LWIPC
	if (LwipcResetEvent(MUSE_SERVER_READY) < 0)
		LOGE("Fail to reset light weight IPC");
#else
	ms_remove_ready_file();
#endif

	retval = muse_server->retval;
	muse_core_fd_close(muse_server->msg_fd);
	muse_core_fd_close(muse_server->data_fd);
	for (idx = 0; idx < MUSE_CHANNEL_MAX; idx++) {
		if (remove(UDS_files[idx]) == MUSE_ERR)
			LOGE("remove %s failed", UDS_files[idx]);
	}

	if (remove(MUSE_DEFAULT_PIDFILE) == -1)
		LOGE("remove %s failed [errno : %d]", MUSE_DEFAULT_PIDFILE, errno);

	for (idx = 0; idx < muse_server->conf->host_cnt; idx++)
		ms_module_deinit(muse_server->module[idx]);

	ms_security_deinit(muse_server->security);
	muse_server->security = NULL;

	ms_system_deinit(muse_server->system);
	muse_server->system = NULL;

	ms_log_deinit(muse_server->log);
	muse_server->log = NULL;

	ms_config_deinit(muse_server->conf);
	muse_server->conf = NULL;

	ms_connection_deinit(muse_server->connection);
	muse_server->connection = NULL;

	muse_core_destroy_fd_table();

	ms_deinit_bufmgr();

	g_mutex_clear(&muse_server->state_lock);

	free(muse_server);
	muse_server = NULL;

	LOGD("Leave");
	return retval;
}

void ms_check_memory(int pid)
{
	int used_pss, memory_threshold;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	ms_connection_t *connection = NULL;

	muse_return_if_fail(muse_server);

	connection = muse_server->connection;
	muse_return_if_fail(connection);

	_ms_lock_state();

	ms_connection_lock(connection);

	used_pss = ms_system_get_memory_usage(pid);

	LOGW("[%d] Proportional set size %d (KByte)", pid, used_pss);

	if (g_queue_is_empty(connection->instance_queue) && ms_is_server_ready()) {
		memory_threshold = ms_config_get_memory_threshold();
		if (used_pss > memory_threshold) {
			ms_set_state(MUSE_SERVER_STATE_IDLE);
			ms_log_process_info(pid);

			snprintf(err_msg, sizeof(err_msg), "[Memory Leak] %d > %d (KByte)", used_pss, memory_threshold);

			LOGE("%s", err_msg);
			ms_connection_unlock(connection);
			_ms_unlock_state();
			ms_respawn(SIGTERM);
		}

		muse_core_remove_all_fd_table();
	}

	ms_connection_unlock(connection);

	_ms_unlock_state();
}

int ms_kill_thread(int signo)
{
	return syscall(__NR_tkill, syscall(__NR_gettid), signo);
}

void ms_new(void)
{
	int fd[MUSE_CHANNEL_MAX];
	int i, j;

	for (i = 0; i < MUSE_CHANNEL_MAX; i++) {
		if (ms_config_is_on_demand() && i == MUSE_CHANNEL_MSG)
			fd[i] = SD_LISTEN_FDS_START;
		else
			fd[i] = _ms_new(i);

		if (!muse_core_fd_is_valid(fd[i])) {
			LOGE("Failed to create socket server %d", i);
			for (j = 0; j < i; j++)
				close(fd[j]);
			return;
		}
	}

	_ms_create_new_server_from_fd(fd, READ | PERSIST);
}

void ms_run(void)
{
	LOGW("Enter");

	muse_return_if_fail(muse_server->main_loop);

	muse_return_if_fail(g_idle_add(_ms_idle_cb, NULL) > 0);

	LOGI("g_main_loop_run");
	g_main_loop_run(muse_server->main_loop);

	LOGW("Leave");
}

void ms_cmd_dispatch(muse_module_h m, muse_module_command_e cmd)
{
	muse_module_cmd_dispatchfunc *cmd_dispatcher = NULL;

	if (m->ch[MUSE_CHANNEL_MSG].dll_handle &&
		g_module_symbol(m->ch[MUSE_CHANNEL_MSG].dll_handle, CMD_DISPATCHER, (gpointer *)&cmd_dispatcher)) {
		if (cmd_dispatcher && cmd_dispatcher[cmd])
			cmd_dispatcher[cmd](m);
	}
}

void ms_respawn(int signo)
{
	pid_t pid = getpid();
	LOGE("send %d process signal %d", (int)pid, signo);

	muse_return_if_fail(muse_server);

	ms_set_state(MUSE_SERVER_STATE_IDLE);
	ms_kill_thread(signo);
}

int ms_get_pid(muse_module_h m)
{
	muse_return_val_if_fail(m, MUSE_ERR);
	return m->pid;
}

void ms_log_process_info(int pid)
{
	muse_core_log_process_thread_info(pid);

	muse_core_log_process_opened_fds(pid);

	muse_core_log_process_cpu_memory(pid);
}

void ms_log_user_group_info(void)
{
	uid_t uid;
	gid_t gid;
	char buffer[MUSE_MSG_MAX_LENGTH];
	struct passwd pwbuf;
	struct passwd *pwbufp;
	struct group gbuf;
	struct group *gbufp;

	uid = getuid();
	gid = getgid();

	getpwuid_r(uid, &pwbuf, buffer, sizeof(buffer), &pwbufp);
	muse_return_if_fail(pwbufp);

	getgrgid_r(gid, &gbuf, buffer, sizeof(buffer), &gbufp);
	muse_return_if_fail(gbufp);

	LOGE("user [%s : %lu] group [%s : %lu]", pwbufp->pw_name, (unsigned long)uid, gbufp->gr_name, (unsigned long)gid);
}

gboolean ms_is_log_enabled(void)
{
	return ms_config_is_log_enabled();
}

gboolean ms_init_bufmgr(void)
{
	LOGD("Enter");

	muse_return_val_if_fail(muse_server, FALSE);

	muse_server->bufmgr = tbm_bufmgr_init(-1);
	if (!muse_server->bufmgr) {
		LOGE("Error - tbm_bufmgr_init");
		ms_log_user_group_info();
		return FALSE;
	}

	LOGD("Leave bufmgr: %p", muse_server->bufmgr);

	return TRUE;
}

void ms_deinit_bufmgr(void)
{
	LOGD("Enter");

	muse_return_if_fail(muse_server);
	muse_return_if_fail(muse_server->bufmgr);

	tbm_bufmgr_deinit(muse_server->bufmgr);

	LOGD("Leave");
}

void ms_cmd_dispatch_foreach_func(gpointer data, gpointer user_data)
{
	muse_module_h m = (muse_module_h)data;
	ms_cmd_dispatcher_info_t *dispatch = (ms_cmd_dispatcher_info_t *)user_data;
	muse_module_command_e cmd;
	muse_external_storage_info_t *storage;

	muse_return_if_fail(muse_server);
	muse_return_if_fail(m);
	muse_return_if_fail(dispatch);

	cmd = dispatch->cmd;

	if (cmd == MUSE_MODULE_COMMAND_EXTERNAL_STORAGE_STATE_CHANGED) {
		storage = &dispatch->storage;
		LOGD("external storage id %d state %d path %s", storage->id, storage->state, storage->path);
		muse_server_set_user_data(m, (void *)storage);
	}

	ms_cmd_dispatch(m, cmd);
	SECURE_LOGI("[%s] %p (%s)", module_cmd[cmd], m, muse_server->conf->host[m->idx]);

	if (cmd == MUSE_MODULE_COMMAND_EXTERNAL_STORAGE_STATE_CHANGED)
		muse_server_set_user_data(m, NULL);
}

void ms_set_state(ms_state_e state)
{
	muse_return_if_fail(muse_server);

	g_mutex_lock(&muse_server->state_lock);
	muse_server->state = state;
	g_mutex_unlock(&muse_server->state_lock);
}

gboolean ms_is_server_ready(void)
{
	muse_return_val_if_fail(muse_server, FALSE);

	return muse_server->state == MUSE_SERVER_STATE_READY;
}

gboolean ms_create_ready_file(void)
{
	LOGD("Enter");
#ifndef MUSE_USE_LWIPC
	int ready_fd;
#endif

#ifdef MUSE_USE_LWIPC
	_ms_wait_event();

	ms_set_state(MUSE_SERVER_STATE_READY);

	if (LwipcEventDone(MUSE_SERVER_READY) < 0) {
		LOGE("Fail to send server ready done event");
		return FALSE;
	}
#else
	ready_fd = creat(MUSE_SERVER_READY, 0644);
	if (muse_core_fd_is_valid(ready_fd)) {
		LOGD("MUSE_SERVER_READY(%s) file was created", MUSE_SERVER_READY);

		close(ready_fd);
		ms_set_state(MUSE_SERVER_STATE_READY);
	} else {
		LOGE("[%d] Fail to create MUSE_SERVER_READY(%s)", ready_fd, MUSE_SERVER_READY);
		return FALSE;
	}
#endif
	LOGD("Leave");

	return TRUE;

}

void ms_remove_ready_file(void)
{
#ifdef MUSE_USE_LWIPC
	if (LwipcResetEvent(MUSE_SERVER_READY) < 0)
		LOGE("Fail to reset light weight IPC");
#else
	if (remove(MUSE_SERVER_READY) == -1)
		LOGE("remove %s failed [errno : %d]", MUSE_SERVER_READY, errno);
#endif
}

