/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"
#include <cynara-client.h>
#include <cynara-session.h>
#include <cynara-creds-socket.h>

static void _ms_security_cynara_log_error(const char *function, int error_code);
static int _ms_security_cynara_new(ms_security_t *security);
static void _ms_security_cynara_free(ms_security_t *security);

static void _ms_security_cynara_log_error(const char *function, int error_code)
{
	char buf[BUFSIZ] = {'\0',};
	cynara_strerror(error_code, buf, sizeof(buf));
	LOGE("[CYNARA] %s function failed with error %d : %s", function, error_code, buf);
}

static int _ms_security_cynara_new(ms_security_t *security)
{
	int ret, error = -1;
	cynara_configuration *p_conf = NULL;

	muse_return_val_if_fail(security, MM_ERROR_INVALID_ARGUMENT);

	if (security->p_cynara) {
		LOGE("[CYNARA] Improper usage of cynara");
		goto end;
	}

	ret = cynara_configuration_create(&p_conf);
	if (ret != CYNARA_API_SUCCESS) {
		_ms_security_cynara_log_error("cynara_configuration_create", ret);
		goto end;
	}

	ret = cynara_configuration_set_cache_size(p_conf, CYNARA_CACHE_SIZE);
	if (ret != CYNARA_API_SUCCESS) {
		_ms_security_cynara_log_error("cynara_configuration_set_cache_size", ret);
		goto end;
	}

	ret = cynara_initialize((cynara **)&security->p_cynara, p_conf);
	if (ret != CYNARA_API_SUCCESS) {
		_ms_security_cynara_log_error("cynara_initialize", ret);
		goto end;
	}

	error = 0;

end:
	cynara_configuration_destroy(p_conf);
	return error;
}

static void _ms_security_cynara_free(ms_security_t *security)
{
	muse_return_if_fail(security);

	if (security->p_cynara) {
		cynara_finish((cynara *)security->p_cynara);
		security->p_cynara = NULL;
	}
}

void ms_security_init(ms_security_t *security)
{
	LOGD("Enter");

	muse_return_if_fail(security);

	_ms_security_cynara_new(security);

	LOGD("Leave");
}

void ms_security_deinit(ms_security_t *security)
{
	LOGD("Enter");

	muse_return_if_fail(security);

	_ms_security_cynara_free(security);

	free(security);
	LOGD("Leave");
}

bool ms_security_is_privilege_granted(ms_security_t *security, int fd, const char *privilege)
{
	int ret = 0;
	pid_t pid = 0;
	char *user = NULL;
	char *client = NULL;
	char *session = NULL;

	muse_return_val_if_fail(security, false);

	ret = cynara_creds_socket_get_user(fd, USER_METHOD_DEFAULT, &user);
	if (ret != CYNARA_API_SUCCESS) {
		_ms_security_cynara_log_error("cynara_creds_socket_get_user", ret);
		goto CLEANUP;
	}

	ret = cynara_creds_socket_get_client(fd, CLIENT_METHOD_DEFAULT, &client);
	if (ret != CYNARA_API_SUCCESS) {
		_ms_security_cynara_log_error("cynara_creds_socket_get_client", ret);
		goto CLEANUP;
	}

	ret = cynara_creds_socket_get_pid(fd, &pid);
	if (ret != CYNARA_API_SUCCESS) {
		_ms_security_cynara_log_error("cynara_creds_socket_get_pid", ret);
		goto CLEANUP;
	}

	session = cynara_session_from_pid(pid);
	if (!session) {
		LOGE("[CYNARA] cynara_session_from_pid() failed");
		ret = CYNARA_API_UNKNOWN_ERROR;
		goto CLEANUP;
	}

	ret = cynara_check((cynara *)security->p_cynara, client, session, user, privilege);
	switch (ret) {
	case CYNARA_API_ACCESS_ALLOWED:
		LOGD("[CYNARA] Check (client = %s, session = %s, user = %s, privilege = %s)"
		"=> access allowed", client, session, user, privilege);
		break;
	case CYNARA_API_ACCESS_DENIED:
		LOGE("[CYNARA] Check (client = %s, session = %s, user = %s, privilege = %s)"
		"=> access denied", client, session, user, privilege);
		break;
	default:
		_ms_security_cynara_log_error("cynara_check", ret);
	}

CLEANUP:
	MUSE_FREE(user);
	MUSE_FREE(session);
	MUSE_FREE(client);

	if (ret == CYNARA_API_ACCESS_ALLOWED)
		return true;
	else
		return false;
}

