/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"
#include <execinfo.h>
#include <dlfcn.h>

struct sigaction ms_segv_old_action;
struct sigaction ms_abrt_old_action;
struct sigaction ms_term_old_action;
struct sigaction ms_bus_old_action;
struct sigaction ms_xcpu_old_action;
struct sigaction ms_pipe_old_action;

static GMutex signal_lock;

static void _ms_signal_handler(int signo);
static void _ms_signal_debug_pid_info(int pid);
static void _ms_signal_backtrace(void *arg);
static void _ms_signal_sigaction(int signo, siginfo_t *si, void *arg);

static void _ms_signal_handler(int signo)
{
	char client_buf[MUSE_MSG_LEN_MAX];

	snprintf(client_buf, sizeof(client_buf), "signo (%d)", signo);
	ms_log_write(client_buf);

	LOGD("signo(%d)", signo);

	ms_set_state(MUSE_SERVER_STATE_IDLE);

	switch (signo) {
	case SIGSEGV:
		sigaction(SIGSEGV, &ms_segv_old_action, NULL);
		break;
	case SIGABRT:
		sigaction(SIGABRT, &ms_abrt_old_action, NULL);
		break;
	case SIGTERM:
		sigaction(SIGTERM, &ms_term_old_action, NULL);
		break;
	case SIGBUS:
		sigaction(SIGBUS, &ms_bus_old_action, NULL);
		break;
	case SIGXCPU:
		sigaction(SIGXCPU, &ms_xcpu_old_action, NULL);
		break;
	default:
		break;
	}

	raise(signo);
}

static void _ms_signal_debug_pid_info(int pid)
{
	if (ms_is_log_enabled())
		ms_log_debug_pid_info(pid);
}

static void _ms_signal_backtrace(void *arg)
{
	void *trace[MUSE_MSG_LEN];
	int tracesize;
	int idx;
	char **strings = NULL;
	ucontext_t *uctxt = NULL;
	Dl_info info;
	char *sym_addr = NULL;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};
	char client_buf[MUSE_MSG_LEN_MAX];

	g_mutex_lock(&signal_lock);

	LOGE("----------BEGIN MUSE DYING MESSAGE----------");

	tracesize = backtrace(trace, MUSE_MSG_LEN);
	if (tracesize < 0) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("backtrace error: %s", err_msg);
	}

	uctxt = (ucontext_t *)arg;

	if (!uctxt) {
		LOGE("Error - null uctxt");
		return;
	}
	#if defined(REG_EIP)
	trace[1] = (void *) uctxt->uc_mcontext.gregs[REG_EIP];
	#elif defined(REG_RIP)
	trace[1] = (void *) uctxt->uc_mcontext.gregs[REG_RIP];
	#endif
	strings = backtrace_symbols(trace, tracesize);
	if (!strings) {
		strerror_r(errno, err_msg, MUSE_MSG_LEN_MAX);
		LOGE("backtrace_symbols error: %s", err_msg);
	} else {
		/* skip the first stack frame because it just points here. */
		for (idx = 1; idx < tracesize; idx++) {
			sym_addr = g_strstr_len(strings[idx], strlen(strings[idx]), "[0x");

			memset(&info, 0, sizeof(info));
			if (sym_addr && dladdr((const void *)strtoul(sym_addr + 1, NULL, 16), &info) && info.dli_sname)
				LOGE("[%u] %s %s", idx - 1, strings[idx], info.dli_sname);
			else
				LOGE("[%u] %s", idx - 1, strings[idx]);
			if (ms_is_log_enabled()) {
				snprintf(client_buf, sizeof(client_buf), "[%u] %s", idx - 1, strings[idx]);
				ms_log_write(client_buf);
			}
		}

		free(strings);
	}

	LOGE("----------END MUSE DYING MESSAGE----------");

	g_mutex_unlock(&signal_lock);
}

static void _ms_signal_sigaction(int signo, siginfo_t *si, void *arg)
{
	muse_return_if_fail(si);
	muse_return_if_fail(arg);

	ms_remove_ready_file();

	ms_config_remove_lockfile();

	_ms_signal_debug_pid_info(si->si_pid);
	_ms_signal_backtrace(arg);
	_ms_signal_handler(signo);
}

void ms_signal_init(void)
{
	struct sigaction action;

	memset(&action, 0, sizeof(struct sigaction));
	sigemptyset(&action.sa_mask);
	action.sa_sigaction = _ms_signal_sigaction;
	action.sa_flags = SA_RESTART | SA_SIGINFO;

	sigaction(SIGSEGV, &action, &ms_segv_old_action);
	sigaction(SIGABRT, &action, &ms_abrt_old_action);
	sigaction(SIGTERM, &action, &ms_term_old_action);
	sigaction(SIGBUS, &action, &ms_bus_old_action);
	sigaction(SIGXCPU, &action, &ms_xcpu_old_action);

	/* Ignore SIGPIPE */
	action.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &action, &ms_pipe_old_action);
}
