/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"
#include <storage.h>
#include <system_info.h>
#include <runtime_info.h>

#define MUSE_STORAGE_EXTERNAL_FEATURE		"http://tizen.org/feature/storage.external"

enum muse_poweroff_type {
	MUSE_POWER_OFF_NONE = 0,
	MUSE_POWER_OFF_POPUP,
	MUSE_POWER_OFF_DIRECT,
	MUSE_POWER_OFF_RESTART,
};

#ifdef MUSE_USE_POWER_OFF_STATE_CHANGE
static int _ms_system_subscribe_poweroff_state_change(ms_system_t *system);
static void _ms_system_unsubscribe_poweroff_state_change(void);
static void _ms_poweroff_state_changed_cb(GDBusConnection *con, const gchar *sender_name, const gchar *object_path,
			const gchar *interface_name, const gchar *signal_name, GVariant *parameters, gpointer user_data);
#endif

#ifdef MUSE_USE_EXTERNAL_STORAGE_STATE_CHANGE
static int _ms_system_subscribe_external_storage_state_change(ms_system_t *system);
static void _ms_system_unsubscribe_external_storage_state_change(void);
static void _ms_external_storage_state_changed_cb(int storage_id, storage_dev_e dev, storage_state_e state, const char *fstype,
			const char *fsuuid, const char *mount_path, bool primary, int flags, void *user_data);
#endif

#ifdef MUSE_USE_RM_READY
static int _ms_system_subscribe_resource_manager_state_change(ms_system_t *system);
static void _ms_system_unsubscribe_resource_manager_state_change(void);
static void _ms_resource_manager_owner_name_changed_cb(GDBusConnection *con, const gchar *sender_name, const gchar *object_path,
			const gchar *interface_name, const gchar *signal_name, GVariant *parameters, gpointer user_data);
#endif

#ifdef MUSE_USE_POWER_OFF_STATE_CHANGE
static int _ms_system_subscribe_poweroff_state_change(ms_system_t *system)
{
	LOGD("Enter");

	muse_return_val_if_fail(system, MM_ERROR_INVALID_ARGUMENT);

	muse_return_val_if_fail(system->connection, MM_ERROR_UNKNOWN);

	system->muse_poweroff_id = g_dbus_connection_signal_subscribe(system->connection, NULL,
		MS_POWER_DBUS_INTERFACE, MS_POWER_DBUS_MSG, NULL, NULL, G_DBUS_SIGNAL_FLAGS_NONE,
		_ms_poweroff_state_changed_cb, NULL, NULL);

	LOGD("Leave");

	return MM_ERROR_NONE;
}

static void _ms_system_unsubscribe_poweroff_state_change(void)
{
	ms_system_t *system = NULL;

	muse_return_if_fail(ms_get_instance());

	system = ms_get_instance()->system;
	muse_return_if_fail(system);
	muse_return_if_fail(system->connection);

	g_dbus_connection_signal_unsubscribe(system->connection, system->muse_poweroff_id);
}

static void _ms_poweroff_state_changed_cb(GDBusConnection *con, const gchar *sender_name, const gchar *object_path,
			const gchar *interface_name, const gchar *signal_name, GVariant *parameters, gpointer user_data)
{
	int val_int = 0;
	GQueue *queue;
	ms_cmd_dispatcher_info_t dispatch;
	ms_system_t *system = NULL;
	ms_connection_t *connection = NULL;

	LOGE("received power off signal");

	muse_return_if_fail(ms_get_instance());
	muse_return_if_fail(object_path);
	muse_return_if_fail(signal_name);
	muse_return_if_fail(parameters);

	system = ms_get_instance()->system;
	muse_return_if_fail(system);

	connection = ms_get_instance()->connection;
	muse_return_if_fail(connection);

	if (g_strcmp0(object_path, MS_POWER_DBUS_PATH) || g_strcmp0(signal_name, MS_POWER_DBUS_MSG)) {
		LOGE("object_path : %s signal_name : %s", object_path, signal_name);
		return;
	}

	g_variant_get(parameters, "(i)", &val_int);

	LOGE("POWER OFF TYPE = %d", val_int);

	/* muse has to restart when MUSE_POWER_OFF_DIRECT and MUSE_POWER_OFF_RESTART */
	muse_return_if_fail(val_int > MUSE_POWER_OFF_POPUP);

	_ms_system_unsubscribe_poweroff_state_change();

	g_mutex_lock(&system->lock);

	ms_connection_lock(connection);

	queue = connection->instance_queue;

	dispatch.cmd = MUSE_MODULE_COMMAND_SHUTDOWN;

	g_queue_foreach(queue, ms_cmd_dispatch_foreach_func, (gpointer)&dispatch);

	ms_connection_unlock(connection);

	g_mutex_unlock(&system->lock);

	exit(EXIT_FAILURE);
}
#endif

#ifdef MUSE_USE_EXTERNAL_STORAGE_STATE_CHANGE
static int _ms_system_subscribe_external_storage_state_change(ms_system_t *system)
{
	LOGD("Enter");

	int ret = storage_set_changed_cb(STORAGE_TYPE_EXTERNAL, _ms_external_storage_state_changed_cb, system);

	if (ret != STORAGE_ERROR_NONE) {
		LOGE("[0x%x] failed to register storage changed callback", ret);
		return MUSE_ERR;
	}

	LOGD("Leave");

	return MM_ERROR_NONE;
}

static void _ms_system_unsubscribe_external_storage_state_change(void)
{
	int ret = storage_unset_changed_cb(STORAGE_TYPE_EXTERNAL, _ms_external_storage_state_changed_cb);
	if (ret != STORAGE_ERROR_NONE)
		LOGE("[0x%x] failed to unregister storage changed callback", ret);
}

static void _ms_external_storage_state_changed_cb(int storage_id, storage_dev_e dev, storage_state_e state, const char *fstype,
			const char *fsuuid, const char *mount_path, bool primary, int flags, void *user_data)
{
	GQueue *queue;
	ms_cmd_dispatcher_info_t dispatch;
	ms_system_t *system = (ms_system_t *)user_data;
	ms_connection_t *connection = NULL;

	muse_return_if_fail(system);
	muse_return_if_fail(ms_get_instance());

	connection = ms_get_instance()->connection;
	muse_return_if_fail(connection);

	switch (state) {
	case STORAGE_STATE_UNMOUNTABLE:
		LOGI("[%d] [%s] STORAGE_STATE_UNMOUNTABLE", state, mount_path);
		break;
	case STORAGE_STATE_REMOVED:
		LOGI("[%d] [%s] STORAGE_STATE_REMOVED", state, mount_path);
		break;
	case STORAGE_STATE_MOUNTED:
		LOGI("[%d] [%s] STORAGE_STATE_MOUNTED", state, mount_path);
		break;
	case STORAGE_STATE_MOUNTED_READ_ONLY:
		LOGI("[%d] [%s] STORAGE_STATE_MOUNTED_READ_ONLY", state, mount_path);
		break;
	default:
		break;
	}

	g_mutex_lock(&system->lock);

	ms_connection_lock(connection);

	queue = connection->instance_queue;

	dispatch.cmd = MUSE_MODULE_COMMAND_EXTERNAL_STORAGE_STATE_CHANGED;

	dispatch.storage.id = storage_id;
	dispatch.storage.state = state;
	dispatch.storage.path = mount_path;

	g_queue_foreach(queue, ms_cmd_dispatch_foreach_func, (gpointer)&dispatch);

	ms_connection_unlock(connection);

	g_mutex_unlock(&system->lock);
}
#endif

#ifdef MUSE_USE_RM_READY
static int _ms_system_subscribe_resource_manager_state_change(ms_system_t *system)
{
	LOGD("Enter");

	muse_return_val_if_fail(system, MM_ERROR_INVALID_ARGUMENT);

	muse_return_val_if_fail(system->connection, MM_ERROR_UNKNOWN);

	system->muse_rm_id = g_dbus_connection_signal_subscribe(system->connection, MS_DBUS_SERVICE,
		MS_DBUS_INTERFACE, MS_RM_DBUS_MSG, NULL, NULL, G_DBUS_SIGNAL_FLAGS_NONE,
		_ms_resource_manager_owner_name_changed_cb, NULL, NULL);
	LOGD("Leave");

	return MM_ERROR_NONE;
}

static void _ms_system_unsubscribe_resource_manager_state_change(void)
{
	ms_system_t *system = NULL;

	muse_return_if_fail(ms_get_instance());

	system = ms_get_instance()->system;

	muse_return_if_fail(system);
	muse_return_if_fail(system->connection);

	g_dbus_connection_signal_unsubscribe(system->connection, system->muse_rm_id);
}

static void _ms_resource_manager_owner_name_changed_cb(GDBusConnection *con, const gchar *sender_name, const gchar *object_path,
			const gchar *interface_name, const gchar *signal_name, GVariant *parameters, gpointer user_data)
{
	GQueue *queue;
	ms_cmd_dispatcher_info_t dispatch;
	ms_system_t *system = NULL;
	ms_connection_t *connection = NULL;
	gchar *owner_name = NULL;
	gchar *old_owner = NULL;
	gchar *new_owner = NULL;

	muse_return_if_fail(ms_get_instance());
	muse_return_if_fail(object_path);
	muse_return_if_fail(signal_name);
	muse_return_if_fail(parameters);

	system = ms_get_instance()->system;
	muse_return_if_fail(system);

	connection = ms_get_instance()->connection;
	muse_return_if_fail(connection);

	if (g_strcmp0(object_path, MS_DBUS_PATH) || g_strcmp0(signal_name, MS_RM_DBUS_MSG))
		return;

	g_variant_get(parameters, "(sss)", &owner_name, &old_owner, &new_owner);

	if (g_strcmp0(owner_name, MS_RM_DBUS_SENDER) || *new_owner == '\0')
		goto out;

	g_mutex_lock(&system->lock);

	ms_connection_lock(connection);

	queue = connection->instance_queue;

	dispatch.cmd = MUSE_MODULE_COMMAND_RESOURCE_MANAGER_SHUTDOWN;

	g_queue_foreach(queue, ms_cmd_dispatch_foreach_func, (gpointer)&dispatch);
	LOGE("resource manager shutdown");

	ms_connection_unlock(connection);

	g_mutex_unlock(&system->lock);
out:
	g_free(owner_name);
	g_free(old_owner);
	g_free(new_owner);
}
#endif

void ms_system_init(ms_system_t *system)
{
	GError *error = NULL;

	muse_return_if_fail(system);

	LOGD("Enter");

	system->connection = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
	if (!system->connection) {
		if (error) {
			LOGE("fail to get gdbus connection (%s)", error->message);
			g_error_free(error);
		}
	}

	system->platform_info_table = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
	muse_return_if_fail(system->platform_info_table);

	g_mutex_init(&system->lock);

	LOGD("Leave");
}

void ms_system_deinit(ms_system_t *system)
{
	muse_return_if_fail(system);
#ifdef MUSE_USE_POWER_OFF_STATE_CHANGE
	_ms_system_unsubscribe_poweroff_state_change();
#endif
#ifdef MUSE_USE_EXTERNAL_STORAGE_STATE_CHANGE
	_ms_system_unsubscribe_external_storage_state_change();
#endif
#ifdef MUSE_USE_RM_READY
	_ms_system_unsubscribe_resource_manager_state_change();
#endif
	if (system->connection)
		g_object_unref(system->connection);

	g_mutex_clear(&system->lock);

	g_hash_table_destroy(system->platform_info_table);

	free(system);
}

int ms_system_get_platform_info(const char *key, bool *value)
{
	int ret = MM_ERROR_NONE;
	gpointer orig_value;
	ms_system_t *system = NULL;

	muse_return_val_if_fail(ms_get_instance(), MM_ERROR_UNKNOWN);

	system = ms_get_instance()->system;

	muse_return_val_if_fail(system, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(key, MM_ERROR_INVALID_ARGUMENT);
	muse_return_val_if_fail(value, MM_ERROR_INVALID_ARGUMENT);

	g_mutex_lock(&system->lock);

	if (g_hash_table_lookup_extended(system->platform_info_table, key, NULL, &orig_value)) {
		*value = GPOINTER_TO_INT(orig_value);
	} else {
		ret = system_info_get_platform_bool(key, value);
		switch (ret) {
		case SYSTEM_INFO_ERROR_NONE: /* Successful */
			g_hash_table_insert(system->platform_info_table, g_strdup(key), GINT_TO_POINTER(*value));
			ret = MM_ERROR_NONE;
			break;
		case SYSTEM_INFO_ERROR_INVALID_PARAMETER: /* Cannot find the key in the model config file */
			ret = MM_ERROR_COMMON_ATTR_NOT_EXIST;
			break;
		case SYSTEM_INFO_ERROR_IO_ERROR: /* An input/output error occurred while reading the value from the model config file */
			ret = MM_ERROR_UNKNOWN;
			break;
		case SYSTEM_INFO_ERROR_PERMISSION_DENIED: /* No permission to use the API */
			ret = MM_ERROR_COMMON_INVALID_PERMISSION;
			break;
		default:
			LOGE("system_info_get_platform_bool returns 0x%x", ret);
			ret = MM_ERROR_UNKNOWN;
			break;
		}
	}

	g_mutex_unlock(&system->lock);

	return ret;
}

void ms_system_subscribe_external_event(ms_system_t *system)
{
#ifdef MUSE_USE_EXTERNAL_STORAGE_STATE_CHANGE
	int ret = MM_ERROR_NONE;
	bool is_external_storage_enabled = false;
#endif
	muse_return_if_fail(system);

#ifdef MUSE_USE_POWER_OFF_STATE_CHANGE
	if (_ms_system_subscribe_poweroff_state_change(system) != MM_ERROR_NONE)
		LOGE("Fail to subscribe poweroff state change");
#endif

#ifdef MUSE_USE_EXTERNAL_STORAGE_STATE_CHANGE
	/* external storage feature is supported as of tizen_5.5 */
	ret = system_info_get_platform_bool(MUSE_STORAGE_EXTERNAL_FEATURE, &is_external_storage_enabled);

	if (ret != MM_ERROR_NONE) /* get the value of mused.conf until tizen_5.0 */
		is_external_storage_enabled = (bool)ms_config_is_external_storage_enabled();

	if (is_external_storage_enabled == true && _ms_system_subscribe_external_storage_state_change(system) != MM_ERROR_NONE)
		LOGE("Fail to subscribe external storage state change");
#endif

#ifdef MUSE_USE_RM_READY
	if (_ms_system_subscribe_resource_manager_state_change(system) != MM_ERROR_NONE)
		LOGE("Fail to subscribe resource manager state change");
#endif
}

int ms_system_get_memory_usage(int pid)
{
	int used_pss = 0;
	process_memory_info_s *info = NULL;
	int ret = runtime_info_get_process_memory_info(&pid, 1, &info);

	if (MM_ERROR_NONE == ret && info)
		used_pss = info->pss;
	else
		LOGE("Fail to get process (%d) memory %s", pid, get_error_message(ret));

	MUSE_FREE(info);

	return used_pss;
}

gboolean ms_system_get_gdbus(GDBusConnection **connection)
{
	LOGD("Enter");

	ms_system_t *system = NULL;

	muse_return_val_if_fail(ms_get_instance(), FALSE);

	system = ms_get_instance()->system;
	muse_return_val_if_fail(system, FALSE);
	muse_return_val_if_fail(system->connection, FALSE);

	*connection = system->connection;

	return TRUE;
}

