/*
 * muse-server
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: YoungHun Kim <yh8004.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "muse_server_private.h"

static gboolean _ms_watchdog_timer_cb(gpointer data);
static gpointer _ms_watchdog_thread(gpointer data);

static gboolean _ms_watchdog_timer_cb(gpointer data)
{
	ms_watchdog_t *watchdog = (ms_watchdog_t *)data;

	muse_return_val_if_fail(watchdog, FALSE);

	g_mutex_lock(&watchdog->lock);
	g_cond_signal(&watchdog->cond);
	g_mutex_unlock(&watchdog->lock);

	return TRUE;
}

static gpointer _ms_watchdog_thread(gpointer data)
{
	ms_watchdog_t *watchdog = (ms_watchdog_t *)data;
	gint64 end_time = 0;
	guint try_count = 0;
	char err_msg[MUSE_MSG_LEN_MAX] = {'\0',};

	muse_return_val_if_fail(watchdog, NULL);

	LOGW("watchdog start : %p", watchdog);

	while (watchdog->run) {
		g_mutex_lock(&watchdog->lock);

		end_time = g_get_monotonic_time() + G_TIME_SPAN_SECOND * MS_WATCHDOG_CHECK_PERIOD;
		if (g_cond_wait_until(&watchdog->cond, &watchdog->lock, end_time)) {
			try_count = 0;
		} else {
			try_count++;
			LOGW("[%d / %d] timeout", try_count, MS_WATCHDOG_CHECK_COUNT);
		}

		if (try_count >= MS_WATCHDOG_CHECK_COUNT) {
			snprintf(err_msg, sizeof(err_msg), "NO RESPONSE OF MAINLOOP");
			LOGE("%s", err_msg);
			g_mutex_unlock(&watchdog->lock);
			ms_kill_thread(SIGABRT);
		}

		g_mutex_unlock(&watchdog->lock);
	}

	LOGW("Leave");

	return NULL;
}

int ms_watchdog_init(ms_watchdog_t *watchdog)
{
	muse_return_val_if_fail(watchdog, MM_ERROR_INVALID_ARGUMENT);

	g_mutex_init(&watchdog->lock);
	g_cond_init(&watchdog->cond);
	watchdog->run = TRUE;

	return MM_ERROR_NONE;
}

int ms_watchdog_deinit(ms_watchdog_t *watchdog)
{
	muse_return_val_if_fail(watchdog, MM_ERROR_INVALID_ARGUMENT);

	g_mutex_clear(&watchdog->lock);
	g_cond_clear(&watchdog->cond);

	return MM_ERROR_NONE;
}

gboolean ms_watchdog_attach(ms_watchdog_t *watchdog)
{
	muse_return_val_if_fail(watchdog, FALSE);

	watchdog->thread = g_thread_try_new("muse_server_watchdog",
										_ms_watchdog_thread,
										(gpointer)watchdog,
										NULL);
	if (watchdog->thread) {
		watchdog->timer_id = g_timeout_add_seconds(MS_WATCHDOG_TIMER_PERIOD,
												_ms_watchdog_timer_cb,
												(gpointer)watchdog);
		return TRUE;
	}

	return FALSE;
}

int ms_watchdog_detach(ms_watchdog_t *watchdog)
{
	gboolean rm_success = FALSE;

	muse_return_val_if_fail(watchdog, MM_ERROR_INVALID_ARGUMENT);

	g_mutex_lock(&watchdog->lock);
	watchdog->run = FALSE;
	g_cond_signal(&watchdog->cond);
	g_mutex_unlock(&watchdog->lock);

	LOGW("join watchdog thread - start");

	g_thread_join(watchdog->thread);
	watchdog->thread = NULL;

	LOGW("join watchdog thread - done");

	rm_success = g_source_remove(watchdog->timer_id);
	if (!rm_success) {
		LOGE("Failed to remove %d", watchdog->timer_id);
		return MM_ERROR_UNKNOWN;
	}

	return MM_ERROR_NONE;
}
