/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "muse_gtest_json.h"
#include "muse_core.h"

static gint _read_file(gchar *file_path, gchar **buf);
static void *_json_thread_function(gpointer buf);

static gint _read_file(gchar *file_path, gchar **buf)
{
	FILE *fp = NULL;
	gchar* tbuf = NULL;

	long file_size = 0;

	if (!file_path) {
		LOGE("file path is null");
		return -1;
	}

	if (!buf) {
		LOGE("buf should not be NULL");
		return -1;
	}

	fp = fopen(file_path, "r");

	if (!fp) {
		LOGE("fp error");
		return -1;
	}

	fseek(fp, 0, SEEK_END);
	file_size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	tbuf = (gchar*)g_malloc0(file_size);
	if (!tbuf) {
		LOGE("Memory allocation failed");
		goto error;
	}

	if (!fgets(tbuf, file_size, fp)) {
		LOGE("Read error");
		goto error;
	}

	*buf = tbuf;

	return file_size;

error:
	g_free(tbuf);
	fclose(fp);

	return -1;
}

static void *_json_thread_function(gpointer buf)
{
	gint *key_request_name = NULL;
	muse_core_msg_parse_err_e err = MUSE_MSG_PARSE_ERROR_NONE;
	gint api, idx, arr_len, url_len;
	void *jobj = NULL;
	gchar *str = NULL;

	if (!buf)
		return GINT_TO_POINTER(FALSE);

	jobj = muse_core_msg_object_new((gchar *)buf, NULL, &err);

	if (!jobj) {
		LOGE("failed to get msg object. jobj err : %d", err);
		goto out;
	}

	if (!muse_core_msg_object_get_value("api", jobj, MUSE_TYPE_INT, &api)) {
		LOGE("failed to get api");
		goto out;
	}
	LOGD("api (gint) : %d\n", api);

	if (!muse_core_msg_object_get_value("key_request_length", jobj, MUSE_TYPE_INT, &arr_len)) {
		LOGE("failed to get length");
		goto out;
	}
	LOGD("key_request_length (gint) : %d\n", arr_len);

	if (!muse_core_msg_object_get_value("server_url_length", jobj, MUSE_TYPE_INT, &url_len)) {
		LOGE("failed to get length");
		goto out;
	}
	LOGD("server_url_length (gint) : %d\n", url_len);

	str = (gchar *)g_try_new0(gchar, MUSE_MSG_LEN_MAX);
	if (!str) {
		LOGE("str is null");
		goto out;
	}

	if (!muse_core_msg_object_get_value("server_url_name", jobj, MUSE_TYPE_STRING, str)) {
		LOGE("failed to get length");
		goto out;
	}
	LOGD("server_url_name (string) : %s\n", str);
	g_free(str);

	key_request_name = (gint *)g_try_new0(gint, arr_len);
	if (key_request_name) {
		LOGD("key_request_name (array) : START");
		if (muse_core_msg_object_get_value("key_request_name", jobj, MUSE_TYPE_ARRAY, key_request_name)) {
			for (idx = 0; idx < arr_len; idx++)
				LOGD("%d ", key_request_name[idx]);
		}
		LOGD("key_request_name (array) : END");
		g_free(key_request_name);
	}

	muse_core_msg_object_free(jobj);

	return GINT_TO_POINTER(TRUE);
out:
	g_free(str);

	if (jobj)
		muse_core_msg_object_free(jobj);

	return GINT_TO_POINTER(FALSE);
}

gboolean muse_gtest_json(gchar *file_path)
{
	GThread *thread;
	gchar* buf = NULL;
	gboolean result = FALSE;

	if (-1 == _read_file(file_path, &buf))
		return FALSE;

	thread = g_thread_new(NULL, _json_thread_function, (gpointer)buf);
	result = GPOINTER_TO_INT(g_thread_join(thread));
	g_free(buf);

	return result;
}

gboolean muse_gtest_multi_threaded_json(gchar *file_path)
{
	gint i;
	GPtrArray *threads = NULL;
	GThread *thread = NULL;
	gboolean result = TRUE;
	gpointer ret;
	gchar* buf = NULL;

	if (-1 == _read_file(file_path, &buf))
		return FALSE;

	threads = g_ptr_array_new();

	for (i = 0; i < NTHREADS; i++) {
		thread = g_thread_new(NULL, _json_thread_function, (gpointer)buf);
		g_ptr_array_add(threads, thread);
	}

	for (i = 0; i < NTHREADS; i++) {
		ret = g_thread_join((GThread *)g_ptr_array_index(threads, i));
		result &= GPOINTER_TO_INT(ret);
	}

	g_ptr_array_free(threads, TRUE);
	g_free(buf);

	return result;
}

